# noqa: F541
"""
This module describes how to match a redis command to grammar token based on
regex.

command_nodex: x means node?
command_keys: ends with s means there can be multiple <key>
"""

import logging
from functools import lru_cache

from prompt_toolkit.contrib.regular_languages.compiler import compile
from .commands import command2syntax

logger = logging.getLogger(__name__)
CONST = {
    "failoverchoice": "TAKEOVER FORCE",
    "withscores": "WITHSCORES",
    "withvalues_const": "WITHVALUES",
    "limit": "LIMIT",
    "expiration": "EX PX",
    "exat_const": "EXAT",
    "pxat_const": "PXAT",
    "condition": "NX XX",
    "keepttl": "KEEPTTL",
    "operation": "AND OR XOR NOT",
    "changed": "CH",
    "incr": "INCR",
    "resetchoice": "HARD SOFT",
    "match": "MATCH",
    "count_const": "COUNT",
    "const_store": "STORE",
    "const_storedist": "STOREDIST",
    "type_const": "TYPE",
    "type": "string list set zset hash stream",
    "position_choice": "BEFORE AFTER",
    "error": "TIMEOUT ERROR",
    "async": "ASYNC SYNC",
    "conntype": "NORMAL MASTER REPLICA PUBSUB",
    "samples": "SAMPLES",
    "slotsubcmd": "IMPORTING MIGRATING NODE STABLE",
    "weights_const": "WEIGHTS",
    "aggregate_const": "AGGREGATE",
    "aggregate": "SUM MIN MAX",
    "slowlogsub": "LEN RESET GET",
    "shutdown": "SAVE NOSAVE",
    "switch": "ON OFF SKIP",
    "on_off": "ON OFF",
    "const_id": "ID",
    "addr": "ADDR",
    "laddr": "LADDR",
    "skipme": "SKIPME",
    "yes": "YES NO",
    "migratechoice": "COPY REPLACE",
    "auth": "AUTH",
    "const_keys": "KEYS",
    "object": "REFCOUNT ENCODING IDLETIME FREQ HELP",
    "subrestore": "REPLACE ABSTTL IDLETIME FREQ",
    "distunit": "m km ft mi",
    "geochoice": "WITHCOORD WITHDIST WITHHASH",
    "order": "ASC DESC",
    "pubsubcmd": "CHANNELS NUMSUB NUMPAT",
    "scriptdebug": "YES NO SYNC",
    "help": "HELP",
    "stream": "STREAM",
    "streams": "STREAMS",
    "stream_create": "CREATE",
    "stream_setid": "SETID",
    "stream_destroy": "DESTROY",
    "stream_delconsumer": "DELCONSUMER",
    "stream_consumers": "CONSUMERS",
    "stream_groups": "GROUPS",
    "stream_group": "GROUP",
    "maxlen": "MAXLEN",
    "idle": "IDLE",
    "time": "TIME",
    "retrycount": "RETRYCOUNT",
    "force": "FORCE",
    "justid": "JUSTID",
    "block": "BLOCK",
    "noack": "NOACK",
    "get": "GET",
    "set": "SET",
    "incrby": "INCRBY",
    "overflow": "OVERFLOW",
    "overflow_option": "WRAP SAT FAIL",
    "version": "VERSION",
    "schedule": "SCHEDULE",
    "graphevent": (
        "ACTIVE-DEFRAG-CYCLE "
        "AOF-FSYNC-ALWAYS "
        "AOF-STAT "
        "AOF-REWRITE-DIFF-WRITE "
        "AOF-RENAME "
        "AOF-WRITE "
        "AOF-WRITE-ACTIVE-CHILD "
        "AOF-WRITE-ALONE "
        "AOF-WRITE-PENDING-FSYNC "
        "COMMAND "
        "EXPIRE-CYCLE "
        "EVICTION-CYCLE "
        "EVICTION-DEL "
        "FAST-COMMAND "
        "FORK "
        "RDB-UNLINK-TEMP-FILE"
    ),
    "section": (
        "SERVER "
        "CLIENTS "
        "MEMORY "
        "PERSISTENCE "
        "STATS "
        "REPLICATION "
        "CPU "
        "COMMANDSTATS "
        "CLUSTER "
        "KEYSPACE "
        "ALL "
        "DEFAULT "
    ),
    "redirect_const": "REDIRECT",
    "prefix_const": "PREFIX",
    "bcast_const": "BCAST",
    "optin_const": "OPTIN",
    "optout_const": "OPTOUT",
    "noloop_const": "NOLOOP",
    "reset_const": "RESET",
    "const_user": "USER",
    "full_const": "FULL",
    "str_algo": "LCS",
    "len_const": "LEN",
    "idx_const": "IDX",
    "minmatchlen_const": "MINMATCHLEN",
    "withmatchlen_const": "WITHMATCHLEN",
    "strings_const": "STRINGS",
    "rank_const": "RANK",
    "lr_const": "LEFT RIGHT",
    "pause_type": "WRITE ALL",
    "db_const": "DB",
    "replace_const": "REPLACE",
    "to_const": "TO",
    "timeout_const": "TIMEOUT",
    "abort_const": "ABORT",
}


def c(const_name):
    const_values = CONST[const_name].split()
    uppers = [x.lower() for x in const_values]
    const_values.extend(uppers)
    return "|".join(const_values)


VALID_TOKEN = r"""(
("([^"]|\\")*?")     |# with double quotes
('([^']|\\')*?')     |# with single quotes
([^\s"]+)            # without quotes
)"""
PATTERN = rf"(?P<pattern>{VALID_TOKEN})"
VALID_SLOT = r"\d+"  # TODO add range? max value:16384
VALID_NODE = r"\w+"
NUM = r"\d+"
NNUM = r"-?\+?\(?\[?(\d+|inf)"  # number cloud be negative
_FLOAT = r"-?(\d|\.|e)+"
DOUBLE = r"\d*(\.\d+)?"
LEXNUM = r"(\[\w+)|(\(\w+)|(\+)|(-)"

SLOT = rf"(?P<slot>{VALID_SLOT})"
SLOTS = rf"(?P<slots>{VALID_SLOT}(\s+{VALID_SLOT})*)"
NODE = rf"(?P<node>{VALID_NODE})"
KEY = rf"(?P<key>{VALID_TOKEN})"
KEYS = rf"(?P<keys>{VALID_TOKEN}(\s+{VALID_TOKEN})*)"
PREFIX = rf"(?P<prefix>{VALID_TOKEN})"
PREFIXES = rf"(?P<prefixes>{VALID_TOKEN}(\s+{VALID_TOKEN})*?)"
DESTINATION = rf"(?P<destination>{VALID_TOKEN})"
NEWKEY = rf"(?P<newkey>{VALID_TOKEN})"
VALUE = rf"(?P<value>{VALID_TOKEN})"
VALUES = rf"(?P<values>{VALID_TOKEN}(\s+{VALID_TOKEN})*)"
ELEMENT = rf"(?P<element>{VALID_TOKEN})"  # element for list
FIELDS = rf"(?P<fields>{VALID_TOKEN}(\s+{VALID_TOKEN})*)"
FIELD = rf"(?P<field>{VALID_TOKEN})"
SFIELD = rf"(?P<sfield>{VALID_TOKEN})"
SVALUE = rf"(?P<svalue>{VALID_TOKEN})"
MEMBER = rf"(?P<member>{VALID_TOKEN})"
MEMBERS = rf"(?P<members>{VALID_TOKEN}(\s+{VALID_TOKEN})*)"
COUNT = rf"(?P<count>{NNUM})"
LEN = rf"(?P<len>{NNUM})"
RANK = rf"(?P<rank>{NNUM})"
VERSION_NUM = rf"(?P<version_num>{NUM})"
MESSAGE = rf"(?P<message>{VALID_TOKEN})"
CHANNEL = rf"(?P<channel>{VALID_TOKEN})"
GROUP = rf"(?P<group>{VALID_TOKEN})"
CONSUMER = rf"(?P<consumer>{VALID_TOKEN})"
CATEGORYNAME = rf"(?P<categoryname>{VALID_TOKEN})"
USERNAME = rf"(?P<username>{VALID_TOKEN})"
RULE = rf"(?P<rule>{VALID_TOKEN})"
BIT = r"(?P<bit>0|1)"
FLOAT = rf"(?P<float>{_FLOAT})"
LONGITUDE = rf"(?P<longitude>{_FLOAT})"
LATITUDE = rf"(?P<latitude>{_FLOAT})"
CURSOR = rf"(?P<cursor>{NUM})"
PARAMETER = rf"(?P<parameter>{VALID_TOKEN})"
PARAMETERS = rf"(?P<parameters>{VALID_TOKEN}(\s+{VALID_TOKEN})*)"
DOUBLE_LUA = r'(?P<double_lua>[^"]*)'
SINGLE_LUA = r"(?P<single_lua>[^']*)"
INTTYPE = r"(?P<inttype>(i|u)\d+)"
# IP re copied from:
# https://www.regular-expressions.info/ip.html
IP = r"""(?P<ip>(25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])\.
               (25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])\.
               (25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])\.
               (25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9]))"""
# Port re copied from:
# https://stackoverflow.com/questions/12968093/regex-to-validate-port-number
# pompt_toolkit limit: Exception: {4}-style repetition not yet supported
PORT = r"(?P<port>[1-9]|[1-5]?\d\d\d?\d?|6[1-4][0-9]\d\d\d|65[1-4]\d\d|655[1-2][0-9]|6553[1-5])"
EPOCH = rf"(?P<epoch>{NUM})"
PASSWORD = rf"(?P<password>{VALID_TOKEN})"
REPLICATIONID = rf"(?P<replicationid>{VALID_TOKEN})"
INDEX = r"(?P<index>(1[0-5]|\d))"
CLIENTID = rf"(?P<clientid>{NUM})"
CLIENTIDS = rf"(?P<clientids>{NUM}(\s+{NUM})*)"

SECOND = rf"(?P<second>{NUM})"
TIMESTAMP = r"(?P<timestamp>[T\d:>+*\-\$]+)"
# TODO test lexer & completer for multi spaces in command
# For now, redis command can have one space at most
COMMAND = r"(\s*  (?P<command>[\w -]+))"
MILLISECOND = rf"(?P<millisecond>{NUM})"
TIMESTAMPMS = r"(?P<timestampms>[T\d:>+*\-\$]+)"
ANY = r"(?P<any>.*)"  # TODO deleted
START = rf"(?P<start>{NNUM})"
END = rf"(?P<end>{NNUM})"

# for stream ids, special ids include:  -, +, $, > and *
# please see:
# https://redis.io/topics/streams-intro#special-ids-in-the-streams-api
# stream id, DO NOT use r"" here, or the \+ will be two string
# NOTE: if miss the outer (), multi IDS won't work.
STREAM_ID = r"(?P<stream_id>[T\d:>+*\-\$]+)"

DELTA = rf"(?P<delta>{NNUM})"
OFFSET = rf"(?P<offset>{NUM})"  # string offset, can't be negative
SHARP_OFFSET = rf"(?P<offset>\#?{NUM})"  # for bitfield command
MIN = rf"(?P<min>{NNUM})"
MAX = rf"(?P<max>{NNUM})"
POSITION = rf"(?P<position>{NNUM})"
SCORE = rf"(?P<score>{_FLOAT})"
LEXMIN = rf"(?P<lexmin>{LEXNUM})"
LEXMAX = rf"(?P<lexmax>{LEXNUM})"
WEIGHTS = rf"(?P<weights>{_FLOAT}(\s+{_FLOAT})*)"
IP_PORT = rf"(?P<ip_port>{IP}:{PORT})"
HOST = rf"(?P<host>{VALID_TOKEN})"
MIN = rf"(?P<min>{NNUM})"
MAX = rf"(?P<max>{NNUM})"
POSITION = rf"(?P<position>{NNUM})"
TIMEOUT = rf"(?P<timeout>{DOUBLE})"
SCORE = rf"(?P<score>{_FLOAT})"
LEXMIN = rf"(?P<lexmin>{LEXNUM})"
LEXMAX = rf"(?P<lexmax>{LEXNUM})"
WEIGHTS = rf"(?P<weights>{_FLOAT}(\s+{_FLOAT})*)"
IP_PORT = rf"(?P<ip_port>{IP}:{PORT})"
HOST = rf"(?P<host>{VALID_TOKEN})"

# const choices
FAILOVERCHOICE = rf"(?P<failoverchoice>{c('failoverchoice')})"
WITHSCORES = rf"(?P<withscores>{c('withscores')})"
LIMIT = rf"(?P<limit>{c('limit')})"
EXPIRATION = rf"(?P<expiration>{c('expiration')})"
CONDITION = rf"(?P<condition>{c('condition')})"
OPERATION = rf"(?P<operation>{c('operation')})"
CHANGED = rf"(?P<changed>{c('changed')})"
INCR = rf"(?P<incr>{c('incr')})"
RESETCHOICE = rf"(?P<resetchoice>{c('resetchoice')})"
MATCH = rf"(?P<match>{c('match')})"
COUNT_CONST = rf"(?P<count_const>{c('count_const')})"
TYPE_CONST = rf"(?P<type_const>{c('type_const')})"
TYPE = rf"(?P<type>{c('type')})"
POSITION_CHOICE = rf"(?P<position_choice>{c('position_choice')})"
ERROR = rf"(?P<error>{c('error')})"
ASYNC = rf"(?P<async>{c('async')})"
CONNTYPE = rf"(?P<conntype>{c('conntype')})"
SAMPLES = rf"(?P<samples>{c('samples')})"
SLOTSUBCMD = rf"(?P<slotsubcmd>{c('slotsubcmd')})"
WEIGHTS_CONST = rf"(?P<weights_const>{c('weights_const')})"
AGGREGATE_CONST = rf"(?P<aggregate_const>{c('aggregate_const')})"
AGGREGATE = rf"(?P<aggregate>{c('aggregate')})"
SLOWLOGSUB = rf"(?P<slowlogsub>{c('slowlogsub')})"
SHUTDOWN = rf"(?P<shutdown>{c('shutdown')})"
SWITCH = rf"(?P<switch>{c('switch')})"
ON_OFF = rf"(?P<on_off>{c('on_off')})"
CONST_ID = rf"(?P<const_id>{c('const_id')})"
CONST_USER = rf"(?P<const_user>{c('const_user')})"
ADDR = rf"(?P<addr>{c('addr')})"
LADDR = rf"(?P<laddr>{c('laddr')})"
SKIPME = rf"(?P<skipme>{c('skipme')})"
YES = rf"(?P<yes>{c('yes')})"
MIGRATECHOICE = rf"(?P<migratechoice>{c('migratechoice')})"
AUTH = rf"(?P<auth>{c('auth')})"
CONST_KEYS = rf"(?P<const_keys>{c('const_keys')})"
OBJECT = rf"(?P<object>{c('object')})"
SUBRESTORE = rf"(?P<subrestore>{c('subrestore')})"
DISTUNIT = rf"(?P<distunit>{c('distunit')})"
GEOCHOICE = rf"(?P<geochoice>{c('geochoice')})"
ORDER = rf"(?P<order>{c('order')})"
CONST_STORE = rf"(?P<const_store>{c('const_store')})"
CONST_STOREDIST = rf"(?P<const_storedist>{c('const_storedist')})"
PUBSUBCMD = rf"(?P<pubsubcmd>{c('pubsubcmd')})"
SCRIPTDEBUG = rf"(?P<scriptdebug>{c('scriptdebug')})"
HELP = rf"(?P<help>{c('help')})"
STREAM = rf"(?P<stream>{c('stream')})"
STREAM_GROUPS = rf"(?P<stream_groups>{c('stream_groups')})"
STREAM_GROUP = rf"(?P<stream_group>{c('stream_group')})"
STREAM_CONSUMERS = rf"(?P<stream_consumers>{c('stream_consumers')})"
STREAM_CREATE = rf"(?P<stream_create>{c('stream_create')})"
STREAM_SETID = rf"(?P<stream_setid>{c('stream_setid')})"
STREAM_DESTROY = rf"(?P<stream_destroy>{c('stream_destroy')})"
STREAM_DELCONSUMER = rf"(?P<stream_delconsumer>{c('stream_delconsumer')})"
MAXLEN = rf"(?P<maxlen>{c('maxlen')})"
APPROXIMATELY = r"(?P<approximately>~)"
IDLE = rf"(?P<idle>{c('idle')})"
TIME = rf"(?P<time>{c('time')})"
RETRYCOUNT = rf"(?P<retrycount>{c('retrycount')})"
FORCE = rf"(?P<force>{c('force')})"
JUSTID = rf"(?P<justid>{c('justid')})"
BLOCK = rf"(?P<block>{c('block')})"
STREAMS = rf"(?P<streams>{c('streams')})"
NOACK = rf"(?P<noack>{c('noack')})"
GET = rf"(?P<get>{c('get')})"
SET = rf"(?P<set>{c('set')})"
INCRBY = rf"(?P<incrby>{c('incrby')})"
OVERFLOW = rf"(?P<overflow>{c('overflow')})"
OVERFLOW_OPTION = rf"(?P<overflow_option>{c('overflow_option')})"
KEEPTTL = rf"(?P<keepttl>{c('keepttl')})"
GRAPHEVENT = rf"(?P<graphevent>{c('graphevent')})"
VERSION = rf"(?P<version>{c('version')})"
SECTION = rf"(?P<section>{c('section')})"
SCHEDULE = rf"(?P<schedule>{c('schedule')})"

REDIRECT_CONST = rf"(?P<redirect_const>{c('redirect_const')})"
PREFIX_CONST = rf"(?P<prefix_const>{c('prefix_const')})"
BCAST_CONST = rf"(?P<bcast_const>{c('bcast_const')})"
OPTIN_CONST = rf"(?P<optin_const>{c('optin_const')})"
OPTOUT_CONST = rf"(?P<optout_const>{c('optout_const')})"
NOLOOP_CONST = rf"(?P<noloop_const>{c('noloop_const')})"

RESET_CONST = rf"(?P<reset_const>{c('reset_const')})"
FULL_CONST = rf"(?P<full_const>{c('full_const')})"

STR_ALGO = rf"(?P<str_algo>{c('str_algo')})"
LEN_CONST = rf"(?P<len_const>{c('len_const')})"
IDX_CONST = rf"(?P<idx_const>{c('idx_const')})"
MINMATCHLEN_CONST = rf"(?P<minmatchlen_const>{c('minmatchlen_const')})"
WITHMATCHLEN_CONST = rf"(?P<withmatchlen_const>{c('withmatchlen_const')})"
STRINGS_CONST = rf"(?P<strings_const>{c('strings_const')})"
RANK_CONST = rf"(?P<rank_const>{c('rank_const')})"

LR_CONST = rf"(?P<lr_const>{c('lr_const')})"
PAUSE_TYPE = rf"(?P<pause_type>{c('pause_type')})"
DB_CONST = rf"(?P<db_const>{c('db_const')})"
REPLACE_CONST = rf"(?P<replace_const>{c('replace_const')})"
TO_CONST = rf"(?P<to_const>{c('to_const')})"
TIMEOUT_CONST = rf"(?P<timeout_const>{c('timeout_const')})"
ABORT_CONST = rf"(?P<abort_const>{c('abort_const')})"
PXAT_CONST = rf"(?P<pxat_const>{c('pxat_const')})"
EXAT_CONST = rf"(?P<exat_const>{c('exat_const')})"
WITHVALUES_CONST = rf"(?P<withvalues_const>{c('withvalues_const')})"

command_grammar = compile(COMMAND)

# Here are the core grammars, those are tokens after ``command``.
# E.g. SET command's syntax is "SET key value"
# Then it's grammar here is r"\s+ key \s+ value \s*", we needn't add `command`
# here because every syntaxes starts with `command` so we will prepend `command`
# in get_command_grammar function.
GRAMMAR = {
    "command_key": rf"\s+ {KEY} \s*",
    "command_pattern": rf"\s+ {PATTERN} \s*",
    "command_command": rf"\s+ {COMMAND} \s*",
    "command_slots": rf"\s+ {SLOTS} \s*",
    "command_node": rf"\s+ {NODE} \s*",
    "command_slot": rf"\s+ {SLOT} \s*",
    "command_failoverchoice": rf"\s+ {FAILOVERCHOICE} \s*",
    "command_resetchoice": rf"\s+ {RESETCHOICE} \s*",
    "command_slot_count": rf"\s+ {SLOT} \s+ {COUNT} \s*",
    "command_key_samples_count": rf"""
        \s+ {KEY} \s+ {SAMPLES} \s+ {COUNT} \s*""",
    "command_key_samples_count_x": rf"\s+ {KEY} (\s+ {SAMPLES} \s+ {COUNT})? \s*",
    "command": r"\s*",
    "command_ip_port": rf"\s+ {IP} \s+ {PORT} \s*",
    "command_epoch": rf"\s+ {EPOCH} \s*",
    "command_yes": rf"\s+ {YES} \s*",
    "command_on_off": rf"\s+ {ON_OFF} \s*",
    "command_sectionx": rf"(\s+ {SECTION})? \s*",
    "command_asyncx": rf"(\s+ {ASYNC})? \s*",
    "command_slot_slotsubcmd_nodex": rf"""
        \s+ {SLOT} \s+ {SLOTSUBCMD} (\s+ {NODE})? \s*""",
    "command_password": rf"\s+ {PASSWORD} \s*",
    "command_usernamex_password": rf"(\s+ {USERNAME})? \s+ {PASSWORD} \s*",
    "command_message": rf"\s+ {MESSAGE} \s*",
    "command_messagex": rf"(\s+{MESSAGE})? \s*",
    "command_index": rf"\s+ {INDEX} \s*",
    "command_index_index": rf"\s+ {INDEX} \s+ {INDEX} \s*",
    "command_client_list": rf"""
        (
            (\s+ {TYPE_CONST} \s+ {CONNTYPE})|
            (\s+ {CONST_ID} \s+ {CLIENTIDS})
        )*
    \s*""",
    "command_clientid_errorx": rf"\s+ {CLIENTID} (\s+ {ERROR})? \s*",
    "command_keys": rf"\s+ {KEYS} \s*",
    "command_key_value": rf"\s+ {KEY} \s+ {VALUE} \s*",
    "command_parameter_value": rf"\s+ {PARAMETER} \s+ {VALUE} \s*",
    "command_parameter_values": rf"(\s+ {PARAMETER} \s+ {VALUE})+ \s*",
    "command_parameter": rf"\s+ {PARAMETER} \s+ {VALUE} \s*",
    "command_parameters": rf"\s+ {PARAMETERS} \s*",
    "command_parametersx": rf"(\s+ {PARAMETERS})? \s*",
    "command_value": rf"\s+ {VALUE} \s*",
    "command_key_second": rf"\s+ {KEY} \s+ {SECOND} \s*",
    "command_key_timestamp": rf"\s+ {KEY} \s+ {TIMESTAMP} \s*",
    "command_key_index": rf"\s+ {KEY} \s+ {INDEX} \s*",
    "command_key_millisecond": rf"\s+ {KEY} \s+ {MILLISECOND} \s*",
    "command_key_timestampms": rf"\s+ {KEY} \s+ {TIMESTAMPMS} \s*",
    "command_key_newkey": rf"\s+ {KEY} \s+ {NEWKEY} \s*",
    "command_newkey_keys": rf"\s+ {NEWKEY} \s+ {KEYS} \s*",
    "command_key_newkey_timeout": rf"\s+ {KEY} \s+ {NEWKEY} \s+ {TIMEOUT} \s*",
    "command_keys_timeout": rf"\s+ {KEYS} \s+ {TIMEOUT} \s*",
    "command_count_timeout": rf"\s+ {COUNT} \s+ {TIMEOUT} \s*",
    "command_pause": rf"\s+ {TIMEOUT} (\s+ {PAUSE_TYPE})? \s*",
    "command_key_positionchoice_pivot_value": rf"""
        \s+ {KEY} \s+ {POSITION_CHOICE} \s+ {VALUE} \s+ {VALUE} \s*""",
    "command_pass": rf"\s+ {ANY} \s*",
    "command_any": rf"\s+ {ANY} \s*",
    "command_set": rf"""
        \s+ {KEY} \s+ {VALUE}
        (
            (\s+ {EXPIRATION} \s+ {MILLISECOND})|
            (\s+ {EXAT_CONST} \s+ {TIMESTAMP})|
            (\s+ {PXAT_CONST} \s+ {TIMESTAMPMS})|
            (\s+ {CONDITION})|
            (\s+ {GET})|
            (\s+ {KEEPTTL})
        )*
        \s*""",
    "command_key_start_end_x": rf"\s+ {KEY} (\s+ {START} \s+ {END})? \s*",
    "command_key_start_end": rf"\s+ {KEY} \s+ {START} \s+ {END} \s*",
    "command_key_delta": rf"\s+ {KEY} \s+ {DELTA} \s*",
    "command_key_offset_value": rf"\s+ {KEY} \s+ {OFFSET} \s+ {VALUE} \s*",
    "command_key_field_value": rf"\s+ {KEY} (\s+ {FIELD} \s+ {VALUE})+ \s*",
    "command_key_offset_bit": rf"\s+ {KEY} \s+ {OFFSET} \s+ {BIT} \s*",
    "command_key_offset": rf"\s+ {KEY} \s+ {OFFSET} \s*",
    "command_key_position": rf"\s+ {KEY} \s+ {POSITION} \s*",
    "command_key_position_value": rf"\s+ {KEY} \s+ {POSITION} \s+ {VALUE} \s*",
    "command_key_second_value": rf"\s+ {KEY} \s+ {SECOND} \s+ {VALUE} \s*",
    "command_key_float": rf"\s+ {KEY} \s+ {FLOAT} \s*",
    "command_key_valuess": rf"(\s+ {KEY} \s+ {VALUE})+ \s*",
    "command_key_values": rf"\s+ {KEY} \s+ {VALUES} \s*",
    "command_key_millisecond_value": rf"\s+ {KEY} \s+ {MILLISECOND} \s+ {VALUE} \s*",
    "command_operation_key_keys": rf"\s+ {OPERATION} \s+ {KEY} \s+ {KEYS} \s*",
    "command_key_bit_start_end": rf"\s+ {KEY} \s+ {BIT} (\s+ {START})? (\s+ {END})? \s*",
    "command_key_members": rf"\s+ {KEY} \s+ {MEMBERS} \s*",
    "command_geodist": rf"\s+ {KEY} \s+ {MEMBER} \s+ {MEMBER} (\s+ {DISTUNIT})? \s*",
    "command_key_longitude_latitude_members": rf"""
        \s+ {KEY}
        (\s+ {CONDITION})?
        (\s+ {CHANGED})?
        (\s+ {LONGITUDE} \s+ {LATITUDE} \s {MEMBER})+
    \s*""",
    "command_destination_keys": rf"\s+ {DESTINATION} \s+ {KEYS} \s*",
    "command_object_key": rf"\s+ {OBJECT} \s+ {KEY} \s*",
    "command_key_member": rf"\s+ {KEY} \s+ {MEMBER} \s*",
    "command_key_any": rf"\s+ {KEY} \s+ {ANY} \s*",
    "command_key_key_any": rf"\s+ {KEY} \s+ {KEY} \s+ {ANY} \s*",
    "command_key_newkey_member": rf"\s+ {KEY} \s+ {NEWKEY} \s+ {MEMBER} \s*",
    "command_key_count_x": rf"\s+ {KEY} (\s+ {COUNT})? \s*",
    "command_key_min_max": rf"\s+ {KEY} \s+ {MIN} \s+ {MAX} \s*",
    "command_key_condition_changed_incr_score_members": rf"""
        \s+ {KEY} (\s+ {CONDITION})?
        (\s+ {CHANGED})?
        (\s+ {INCR})?
        (\s+ {SCORE} \s+ {MEMBER})+ \s*""",
    "command_key_float_member": rf"\s+ {KEY} \s+ {FLOAT} \s+ {MEMBER} \s*",
    "command_key_lexmin_lexmax": rf"\s+ {KEY} \s+ {LEXMIN} \s+ {LEXMAX} \s*",
    "command_key_start_end_withscores_x": rf"""
        \s+ {KEY} \s+ {START} \s+ {END} (\s+ {WITHSCORES})? \s*""",
    "command_key_lexmin_lexmax_limit_offset_count": rf"""
        \s+ {KEY} \s+ {LEXMIN} \s+ {LEXMAX}
        (\s+ {LIMIT} \s+ {OFFSET} \s+ {COUNT})? \s*""",
    "command_key_min_max_withscore_x_limit_offset_count_x": rf"""
        \s+ {KEY} \s+ {MIN} \s+ {MAX} (\s+ {WITHSCORES})?
        (\s+ {LIMIT} \s+ {OFFSET} \s+ {COUNT})? \s*""",
    "command_cursor_match_pattern_count_type": rf"""
        \s+ {CURSOR}
        (
            (\s+ {MATCH} \s+ {PATTERN})|
            (\s+ {COUNT_CONST} \s+ {COUNT})|
            (\s+ {TYPE_CONST} \s+ {TYPE})
        )*
        \s*""",
    "command_key_cursor_match_pattern_count": rf"""\s+ {KEY}
        \s+ {CURSOR}
        (
            (\s+ {MATCH} \s+ {PATTERN})|
            (\s+ {COUNT_CONST} \s+ {COUNT})
        )*
        \s*""",
    "command_key_fields": rf"\s+ {KEY} \s+ {FIELDS} \s*",
    "command_key_field": rf"\s+ {KEY} \s+ {FIELD} \s*",
    "command_key_field_delta": rf"\s+ {KEY} \s+ {FIELD} \s+ {DELTA} \s*",
    "command_key_field_float": rf"\s+ {KEY} \s+ {FIELD} \s+ {FLOAT} \s*",
    "command_key_fieldvalues": rf"\s+ {KEY} (\s+ {FIELD} \s+ {VALUE})+ \s*",
    "command_slowlog": rf"\s+ {SLOWLOGSUB} \s+ {NUM} \s*",
    "command_switch": rf"\s+ {SWITCH} \s*",
    "command_schedulex": rf"(\s+ {SCHEDULE})? \s*",
    "command_clientkill": rf"""
        (
            (\s+ {IP_PORT})|
            (\s+ {ADDR} \s+ {IP_PORT})|
            (\s+ {LADDR} \s+ {IP_PORT})|
            (\s+ {CONST_ID} \s+ {CLIENTID})|
            (\s+ {TYPE_CONST} \s+ {CONNTYPE})|
            (\s+ {CONST_USER} \s+ {USERNAME})|
            (\s+ {SKIPME} \s+ {YES})
        )+ \s*""",
    "command_migrate": rf"""
        \s+ {HOST} \s+ {PORT}
        \s+ {KEY} \s+ {INDEX} \s+ {TIMEOUT}
        (\s+ {MIGRATECHOICE})?
        (
            (\s+ {AUTH} \s+ {PASSWORD})|
            (\s+ {AUTH} \s+ {USERNAME} \s+ {PASSWORD})
        )?
        (\s+ {CONST_KEYS} \s+ {KEYS})?
    \s*""",
    "command_restore": rf"""\s+ {KEY}
        \s+ {TIMEOUT} \s+ {VALUE} (\s+ {SUBRESTORE} \s+ {SECOND})? \s*""",
    "command_pubsubcmd_channels": rf"\s+ {PUBSUBCMD} (\s+ {CHANNEL})+ \s*",
    "command_pubsubcmd_channelsx": rf"\s+ {PUBSUBCMD} (\s+ {CHANNEL})* \s*",
    "command_channel_message": rf"\s+ {CHANNEL} \s+ {MESSAGE} \s*",
    "command_channels": rf"(\s+ {CHANNEL})+ \s*",
    "command_channelsx": rf"(\s+ {CHANNEL})* \s*",
    "command_lua_any": rf"""(\s+"{DOUBLE_LUA}")? (\s+'{SINGLE_LUA}')? \s+ {ANY} \s*""",
    "command_scriptdebug": rf"\s+ {SCRIPTDEBUG} \s*",
    "command_shutdown": rf"\s+ {SHUTDOWN} \s*",
    "command_key_start_end_countx": rf"""\s+ {KEY}
        \s+ {STREAM_ID}
        \s+ {STREAM_ID}
        (\s+ {COUNT_CONST} \s+ {COUNT})?
        \s*""",
    "command_xgroup": rf"""
        (
            (\s+ {STREAM_CREATE} \s+ {KEY} \s+ {GROUP} \s+ {STREAM_ID})|
            (\s+ {STREAM_SETID} \s+ {KEY} \s+ {GROUP} \s+ {STREAM_ID})|
            (\s+ {STREAM_DESTROY} \s+ {KEY} \s+ {GROUP})|
            (\s+ {STREAM_DELCONSUMER} \s+ {KEY} \s+ {GROUP} \s+ {CONSUMER})|
            (\s+ {HELP})
        )
        \s*""",
    "command_key_group_ids": rf"""
        \s+ {KEY} \s+ {GROUP} (\s+ {STREAM_ID})+ \s*""",
    "command_key_ids": rf"""
        \s+ {KEY} (\s+ {STREAM_ID})+ \s*""",
    "command_xinfo": rf"""
        (
            (\s+ {STREAM_CONSUMERS} \s+ {KEY} \s+ {GROUP})|
            (\s+ {STREAM_GROUPS} \s+ {KEY})|
            (\s+ {STREAM} \s+ {KEY}
                (\s+ {FULL_CONST})?
                (\s+ {COUNT_CONST} \s+ {COUNT})?
            )|
            (\s+ {HELP})
        )
        \s*""",
    "command_xpending": rf"""
        \s+ {KEY} \s+ {GROUP}
        (\s+ {STREAM_ID} \s+ {STREAM_ID} \s+ {COUNT})?
        (\s+ {CONSUMER})?
        \s*""",
    "command_xadd": rf"""
        \s+ {KEY}
        (\s+ {MAXLEN} (\s+ {APPROXIMATELY})? \s+ {COUNT})?
        \s+ {STREAM_ID}
        (\s+ {SFIELD} \s+ {SVALUE})+ \s*""",
    "command_key_maxlen": rf"""
        \s+ {KEY} \s+ {MAXLEN} (\s+ {APPROXIMATELY})? \s+ {COUNT}
        \s*""",
    "command_xclaim": rf"""
        \s+ {KEY} \s+ {GROUP} \s+ {CONSUMER} \s+ {MILLISECOND}
        (\s+ {STREAM_ID})+
        (\s+ {IDLE} \s+ {MILLISECOND})?
        (\s+ {TIME} \s+ {TIMESTAMP})?
        (\s+ {RETRYCOUNT} \s+ {COUNT})?
        (\s+ {FORCE})?
        (\s+ {JUSTID})?
        \s*""",
    "command_xread": rf"""
        (\s+ {COUNT_CONST} \s+ {COUNT})?
        (\s+ {BLOCK} \s+ {MILLISECOND})?
        \s+ {STREAMS}
        \s+ {KEYS}
        (\s+ {STREAM_ID})+
        \s*""",
    "command_xreadgroup": rf"""
        \s+ {STREAM_GROUP} \s+ {GROUP} \s+ {CONSUMER}
        (\s+ {COUNT_CONST} \s+ {COUNT})?
        (\s+ {BLOCK} \s+ {MILLISECOND})?
        (\s+ {NOACK})?
        \s+ {STREAMS}
        \s+ {KEYS}
        (\s+ {STREAM_ID})+
        \s*""",
    "command_bitfield": rf"""
        \s+ {KEY}
        (
            (\s+ {GET} \s+ {INTTYPE} \s+ {SHARP_OFFSET})|
            (\s+ {SET} \s+ {INTTYPE} \s+ {SHARP_OFFSET} \s+ {VALUE})|
            (\s+ {INCRBY} \s+ {INTTYPE} \s+ {SHARP_OFFSET} \s+ {VALUE})|
            (\s+ {OVERFLOW} \s+ {OVERFLOW_OPTION})
        )+
        \s*""",
    "command_replicationid_offset": rf"\s+ {REPLICATIONID} \s+ {OFFSET} \s*",
    "command_graphevent": rf"\s+ {GRAPHEVENT} \s*",
    "command_graphevents": rf"(\s+ {GRAPHEVENT})* \s*",
    # before redis 5: lolwut 5 1
    # start from redis 6: lolwut VERSION 5 1
    "command_version": rf"(\s+ {VERSION} \s+ {VERSION_NUM})? (\s+ {ANY})? \s*",
    "command_client_tracking": rf"""
        \s+ {ON_OFF}
        (
            (\s+ {REDIRECT_CONST} \s+ {CLIENTID})|
            (\s+ {PREFIX_CONST} \s+ {PREFIXES})|
            (\s+ {BCAST_CONST})|
            (\s+ {OPTIN_CONST})|
            (\s+ {OPTOUT_CONST})|
            (\s+ {NOLOOP_CONST})
        )*
        \s*""",
    "command_categorynamex": rf"(\s+ {CATEGORYNAME})? \s*",
    "command_usernames": rf"(\s+ {USERNAME})+ \s*",
    "command_username": rf"\s+ {USERNAME} \s*",
    "command_count_or_resetx": rf"( (\s+ {COUNT}) | (\s+ {RESET_CONST}) )? \s*",
    "command_username_rules": rf"\s+ {USERNAME} (\s+ {RULE})* \s*",
    "command_username_parameter_parametersx": rf"\s+ {USERNAME} \s+ {PARAMETER} (\s+ {PARAMETERS})? \s*",
    "command_count": rf"(\s+ {COUNT})? \s*",
    "command_stralgo": rf"""
        (
            \s+ {STR_ALGO}
            (
                (\s+ {CONST_KEYS} \s+ {KEYS})|
                (\s+ {STRINGS_CONST} \s+ {VALUES})
            )
            (\s+ {IDX_CONST})?
            (\s+ {LEN_CONST})?
            (\s+ {MINMATCHLEN_CONST} \s+ {LEN})?
            (\s+ {WITHMATCHLEN_CONST})?
        )
    \s*""",
    "command_lpos": rf"""
        \s+ {KEY} \s+ {ELEMENT}
        (
            (\s+ {RANK_CONST} \s+ {RANK})|
            (\s+ {COUNT_CONST} \s+ {COUNT})|
            (\s+ {MAXLEN} \s+ {LEN})
        )*
        \s*""",
    "command_key_key_lr_lr_timeout": rf"""
        \s+ {KEY} \s+ {KEY}
        \s+ {LR_CONST} \s+ {LR_CONST}
        \s+ {TIMEOUT} \s*""",
    "command_key_key_lr_lr": rf"""
        \s+ {KEY} \s+ {KEY}
        \s+ {LR_CONST} \s+ {LR_CONST} \s*""",
    "command_copy": rf"""
        \s+ {KEY} \s+ {KEY}
        (\s+ {DB_CONST} \s+ {INDEX})?
        (\s+ {REPLACE_CONST})?
        \s*""",
    "command_failover": rf"""
        (\s+ {TO_CONST} \s+ {HOST} \s+ {PORT} (\s+ {FORCE})? )?
        (\s+ {ABORT_CONST})?
        (\s+ {TIMEOUT_CONST} \s+ {MILLISECOND})?
        \s*""",
    "command_key_expire": rf"""
        \s+ {KEY}
        (
            (\s+ {EXPIRATION} \s+ {MILLISECOND})|
            (\s+ {PXAT_CONST} \s+ {TIMESTAMPMS})|
            (\s+ {EXAT_CONST} \s+ {TIMESTAMP})
        )?
        \s*""",
    "command_key_count_withvalues": rf"""
        \s+ {KEY}
        (\s+ {COUNT} (\s+ {WITHVALUES_CONST})?)?
        \s*""",
}

pipeline = r"(?P<shellcommand>\|.*)?"


@lru_cache(maxsize=256)
def get_command_grammar(command):
    """
    :param command: command name in upper case. This command must be raw user
        input, otherwise can't match in lexer, cause this command to be invalid;
    """
    syntax_name = command2syntax[" ".join(command.split()).upper()]
    syntax = GRAMMAR.get(syntax_name)

    # If a command is not supported yet, (e.g. command from latest version added
    # by Redis recently, or command from third Redis module.) return a default
    # grammar, so the lexer and completion won't be activated.
    if syntax is None:
        return command_grammar
    # prepend command token for this syntax
    command_allow_multi_spaces = command.replace(r" ", r"\s+")
    syntax = rf"\s* (?P<command>{command_allow_multi_spaces}) " + syntax
    # allow user input pipeline to redirect to shell, like `get json | jq .`
    syntax += pipeline

    logger.info(f"syxtax: {syntax}")

    return compile(syntax)



class IRedisException(Exception):
    pass


class UsageError(IRedisException):
    pass


class InvalidArguments(IRedisException):
    """Invalid argument(s)"""


class NotRedisCommand(IRedisException):
    """Not a Redis command"""


class AmbiguousCommand(IRedisException):
    """Command is not finished, don't it's command's name"""


class NotSupport(IRedisException):
    """IRedis currently not support this."""



import logging

from prompt_toolkit.layout.processors import (
    Processor,
    Transformation,
    TransformationInput,
)

from .exceptions import InvalidArguments, AmbiguousCommand
from .commands import split_command_args

logger = logging.getLogger(__name__)


class UserInputCommand:
    """
    User inputted command in real time.

    ``UpdateBottomProcessor`` update it, and ``BottomToolbar`` read it
    """

    def __init__(self):
        # command will always be upper case
        self.command = None


class UpdateBottomProcessor(Processor):
    """
    Update Footer display text while user input.
    """

    def __init__(self, command_holder, session):
        # processor will call for internal_refresh, when input_text didn't
        # change, don't run
        self.session = session
        self.command_holder = command_holder

    def apply_transformation(
        self, transformation_input: TransformationInput
    ) -> Transformation:
        input_text = transformation_input.document.text
        try:
            command, _ = split_command_args(input_text)
        except (InvalidArguments, AmbiguousCommand):
            self.command_holder.command = None
        else:
            self.command_holder.command = command.upper()

        return Transformation(transformation_input.fragments)


class PasswordProcessor(Processor):
    """
    Processor that turns masks the input. (For passwords.)

    :param char: (string) Character to be used. "*" by default.
    """

    def __init__(self, char: str = "*") -> None:
        self.char = char

    def apply_transformation(self, ti: TransformationInput) -> Transformation:
        input_text = ti.document.text
        default_transformation = Transformation(ti.fragments)
        try:
            command, _ = split_command_args(input_text)
        except (InvalidArguments, AmbiguousCommand):
            return default_transformation

        if command.upper() != "AUTH":
            return default_transformation

        fragments = []
        for style, text, *handler in ti.fragments:
            if style == "class:password":
                fragments.append((style, self.char * len(text), *handler))
            else:
                fragments.append((style, text, *handler))
        return Transformation(fragments)



import logging

from prompt_toolkit.filters import completion_is_selected
from prompt_toolkit.key_binding import KeyBindings

logger = logging.getLogger(__name__)

kb = KeyBindings()


@kb.add("enter", filter=completion_is_selected)
def _(event):
    """Makes the enter key work as the tab key only when showing the menu.
    In other words, don't execute query when enter is pressed in
    the completion dropdown menu, instead close the dropdown menu
    (accept current selection).
    """
    logger.debug("Detected enter key.")

    event.current_buffer.complete_state = None
    b = event.app.current_buffer
    b.complete_state = None



from importlib.resources import as_file, files
import os
import logging

from configobj import ConfigObj, ConfigObjError
from . import data as project_data

# TODO verbose logger to print to stdout
logger = logging.getLogger(__name__)

system_config_file = "/etc/iredisrc"
pwd_config_file = os.path.join(os.getcwd(), ".iredisrc")


class Config:
    """
    Global config, set once on start, then
    become readonly, never change again.

    :param raw: weather write raw bytes to stdout without any
        decoding.
    :param decode: How to decode bytes response.(For display and
        Completers)
        default is None, means show literal bytes. But completers
        will try use utf-8 decoding.
    """

    def __init__(self):
        self.raw = None
        self.completer_max = None
        # show command hint?
        self.newbie_mode = None
        self.rainbow = None
        self.retry_times = 2
        self.socket_keepalive = None
        self.decode = None
        self.no_info = None
        self.bottom_bar = None
        self.shell = None
        self.enable_pager = None
        self.pager = None
        self.verify_ssl = None

        self.warning = True

        self.no_version_reason = None
        self.log_location = None
        self.history_location = None
        self.completion_casing = None
        self.alias_dsn = None
        self.natmap = {}

        # ===bad code===
        # below are not configs, it's global state, it's wrong to write this
        # please do not add more global state.
        # FIXME this should be removed.
        # use client attributes instead.
        # use kwargs in render functions.

        # for transaction render
        self.queued_commands = []
        self.transaction = False
        # display zset withscores?
        self.withscores = False
        self.version = "Unknown"

        self.greetings = True

        self.prompt = None

    def __setter__(self, name, value):
        # for every time start a transaction
        # clear the queued commands first
        if name == "transaction" and value is True:
            self.queued_commands = []
        super().__setattr__(name, value)


config = Config()


def read_config_file(f):
    """Read a config file."""

    if isinstance(f, str):
        f = os.path.expanduser(f)

    try:
        config = ConfigObj(f, interpolation=False, encoding="utf8")
    except ConfigObjError as e:
        logger.error(
            "Unable to parse line {} of config file " "'{}'.".format(e.line_number, f)
        )
        logger.error("Using successfully parsed config values.")
        return e.config
    except OSError as e:
        logger.error(
            "You don't have permission to read " "config file '{}'.".format(e.filename)
        )
        return None

    return config


def load_config_files(iredisrc):
    with as_file(files(project_data).joinpath("iredisrc")) as p:
        config_obj = ConfigObj(str(p))

    for _file in [system_config_file, iredisrc, pwd_config_file]:
        _config = read_config_file(_file)
        if bool(_config) is True:
            config_obj.merge(_config)
            config_obj.filename = _config.filename

    config.raw = config_obj["main"].as_bool("raw")
    config.completer_max = config_obj["main"].as_int("completer_max")
    config.retry_times = config_obj["main"].as_int("retry_times")
    config.newbie_mode = config_obj["main"].as_bool("newbie_mode")
    config.rainbow = config_obj["main"].as_bool("rainbow")
    config.socket_keepalive = config_obj["main"].as_bool("socket_keepalive")
    config.no_info = config_obj["main"].as_bool("no_info")
    config.bottom_bar = config_obj["main"].as_bool("bottom_bar")
    config.warning = config_obj["main"].as_bool("warning")
    config.decode = config_obj["main"]["decode"]
    config.log_location = config_obj["main"]["log_location"]
    config.completion_casing = config_obj["main"]["completion_casing"]
    config.history_location = config_obj["main"]["history_location"]
    config.alias_dsn = config_obj["alias_dsn"]
    config.shell = config_obj["main"].as_bool("shell")
    config.pager = config_obj["main"].get("pager")
    config.enable_pager = config_obj["main"].as_bool("enable_pager")
    config.prompt = config_obj["main"].get("prompt")
    config.greetings = config_obj["main"].as_bool("greetings")

    return config_obj



"""
IRedis client.
"""

import re
import os
import sys
import codecs
import logging
from subprocess import run
from importlib.resources import files
from packaging.version import parse as version_parse

import redis
from prompt_toolkit.shortcuts import clear
from prompt_toolkit.formatted_text import FormattedText
from redis.connection import Connection, SSLConnection, UnixDomainSocketConnection
from redis.exceptions import (
    AuthenticationError,
    ConnectionError,
    MovedError,
    TimeoutError,
    ResponseError,
)

from . import markdown, renders
from .data import commands as commands_data
from .commands import (
    command2callback,
    commands_summary,
    command2syntax,
    groups,
    split_command_args,
    split_unknown_args,
)
from .completers import IRedisCompleter
from .config import config
from .exceptions import NotRedisCommand, InvalidArguments, AmbiguousCommand, NotSupport
from .renders import OutputRender
from .utils import (
    compose_command_syntax,
    nativestr,
    exit,
    convert_formatted_text_to_bytes,
    parse_url,
)
from .warning import confirm_dangerous_command

logger = logging.getLogger(__name__)
CLIENT_COMMANDS = groups["iredis"]


class Client:
    """
    iRedis client, hold a redis-py Client to interact with Redis.
    """

    def __init__(
        self,
        host="127.0.0.1",
        port=6379,
        db=0,
        password=None,
        path=None,
        scheme="redis",
        username=None,
        client_name=None,
        prompt=None,
        verify_ssl=None,
    ):
        self.host = host
        self.port = port
        self.db = db
        self.path = path
        self.username = username
        self.client_name = client_name
        self.scheme = scheme
        self.password = password

        # cli args --prompt will overwrite the prompt in iredisrc config file
        self.prompt = ""
        if config.prompt:
            self.prompt = config.prompt
        if prompt:
            self.prompt = prompt

        self.verify_ssl = verify_ssl or "required"

        self.client_id = None
        self.client_addr = None

        self.build_connection()

        # all command upper case
        self.answer_callbacks = command2callback
        self.set_default_pager(config)
        try:
            self.connection.connect()
        except Exception as e:
            logger.exception("Can not create connection to server")
            print(str(e), file=sys.stderr)
            sys.exit(1)
        if not config.no_info:
            try:
                self.get_server_info()
            except Exception as e:
                logger.warning(f"[After Connection] {str(e)}")
                config.no_version_reason = str(e)
        else:
            config.no_version_reason = "--no-info flag activated"

        if self.prompt and "client_addr" in self.prompt:
            self.client_addr = ":".join(
                str(x) for x in self.connection._sock.getsockname()
            )
        if self.prompt and "client_id" in self.prompt:
            self.client_id = str(self.execute("CLIENT ID"))

        if config.version and re.match(r"([\d\.]+)", config.version):
            self.auth_compat(config.version)

    def build_connection(self):
        """
        create a new connection and replace ``self.connection``
        """
        self.connection = self.create_connection(
            self.host,
            self.port,
            self.db,
            self.password,
            self.path,
            self.scheme,
            self.username,
            self.verify_ssl,
            client_name=self.client_name,
        )

    def create_connection(
        self,
        host=None,
        port=None,
        db=0,
        password=None,
        path=None,
        scheme="redis",
        username=None,
        verify_ssl=None,
        client_name=None,
    ):
        if scheme in ("redis", "rediss"):
            connection_kwargs = {
                "host": host,
                "port": port,
                "db": db,
                "password": password,
                "socket_keepalive": config.socket_keepalive,
                "client_name": client_name,
            }

            # if username is set without setting paswword, password will be ignored
            # But empty password will be properly handled
            if password is not None:
                connection_kwargs["username"] = username

            if scheme == "rediss":
                connection_kwargs["ssl_cert_reqs"] = verify_ssl
                connection_class = SSLConnection
            else:
                connection_class = Connection
        else:
            connection_kwargs = {
                "db": db,
                "password": password,
                "path": path,
                "client_name": client_name,
                "username": username,
            }
            connection_class = UnixDomainSocketConnection

        if config.decode:
            connection_kwargs["encoding"] = config.decode
            connection_kwargs["decode_responses"] = True
            connection_kwargs["encoding_errors"] = "replace"

        logger.debug(
            f"connection_class={connection_class},"
            f" connection_kwargs={connection_kwargs}"
        )

        return connection_class(**connection_kwargs)

    def auth_compat(self, redis_version: str):
        with_username = version_parse(redis_version) >= version_parse("6.0.0")
        if with_username:
            command2syntax["AUTH"] = "command_usernamex_password"
        else:
            command2syntax["AUTH"] = "command_password"

    def set_default_pager(self, config):
        configured_pager = config.pager
        os_environ_pager = os.environ.get("PAGER")

        if configured_pager:
            logger.info('Default pager found in config file: "%s"', configured_pager)
            os.environ["PAGER"] = configured_pager
        elif os_environ_pager:
            logger.info(
                'Default pager found in PAGER environment variable: "%s"',
                os_environ_pager,
            )
            os.environ["PAGER"] = os_environ_pager
        else:
            logger.info("No default pager found in environment. Using os default pager")

        # Set default set of less recommended options, if they are not already set.
        # They are ignored if pager is different than less.
        if not os.environ.get("LESS"):
            os.environ["LESS"] = "-SRXF"

    def get_server_info(self):
        # safe to decode Redis's INFO response
        info_resp = nativestr(self.execute("INFO"))
        version = re.findall(r"redis_version:(.+)\r\n", info_resp)[0]
        logger.debug(f"[Redis Version] {version}")
        config.version = version

    def __str__(self):
        if self.prompt:  # not None and not empty
            return self.prompt.format(
                client_name=self.client_name,
                db=self.db,
                host=self.host,
                path=self.path,
                port=self.port,
                username=self.username,
                client_addr=self.client_addr,
                client_id=self.client_id,
            )

        if self.scheme == "unix":
            prompt = f"redis {self.path}"
        else:
            prompt = f"{self.host}:{self.port}"

        if self.db:
            prompt = f"{prompt}[{self.db}]"

        return f"{prompt}> "

    def client_execute_command(self, command_name, *args):
        command = command_name.upper()
        if command == "HELP":
            yield self.do_help(*args)
        if command == "PEEK":
            yield from self.do_peek(*args)
        if command == "CLEAR":
            clear()
        if command == "EXIT":
            exit()

    def execute(self, *args, **kwargs):
        logger.info(
            f"execute: connection={self.connection} args={args}, kwargs={kwargs}"
        )
        return self.execute_by_connection(self.connection, *args, **kwargs)

    def execute_by_connection(self, connection, command_name, *args, **options):
        """Execute a command and return a parsed response
        Here we retry once for ConnectionError.
        """
        logger.info(
            f"execute by connection: connection={connection}, name={command_name},"
            f" {args}, {options}"
        )
        retry_times = config.retry_times  # FIXME configurable
        last_error = None
        need_refresh_connection = False

        while retry_times >= 0:
            try:
                if need_refresh_connection:
                    print(
                        f"{str(last_error)} retrying... retry left: {retry_times+1}",
                        file=sys.stderr,
                    )
                    connection.disconnect()
                    connection.connect()
                    logger.info(f"New connection created, retry on {connection}.")
                logger.info(f"send_command: {command_name} , {args}")

                connection.send_command(command_name, *args)
                response = connection.read_response()
            except AuthenticationError:
                raise
            except (ConnectionError, TimeoutError) as e:
                logger.warning(f"Connection Error, got {e}, retrying...")
                last_error = e
                retry_times -= 1
                need_refresh_connection = True
            except redis.exceptions.ExecAbortError:
                config.transaction = False
                raise
            except MovedError as e:
                return self.reissue_with_redirect(
                    f"MOVED {str(e)}", command_name, *args, **options
                )
            except ResponseError as e:
                response_message = str(e)
                if response_message.startswith("MOVED"):
                    return self.reissue_with_redirect(
                        response_message, command_name, *args, **options
                    )
                raise e

            except KeyboardInterrupt:
                logger.warning("received KeyboardInterrupt... rebuild connection...")
                connection.disconnect()
                connection.connect()
                print(
                    "KeyboardInterrupt received! User canceled reading response!",
                    file=sys.stderr,
                )
                return None
            else:
                return response
        if last_error:
            raise last_error

    def reissue_with_redirect(self, response, *args, **kwargs):
        """
        For redis cluster, when server response a "MOVE ..." response, we auto-
        redirect to the target node, reissue the original command.

        This feature is not supported for unix socket connection.
        """
        # Redis Cluster only supports database zero.
        _, _, ip_port = response.split(" ")
        ip, port = ip_port.split(":")
        port = int(port)

        if f"{ip}:{port}" in config.natmap:
            ip, port = config.natmap[f"{ip}:{port}"]

        print(response, file=sys.stderr)

        connection = self.create_connection(
            ip,
            port,
            username=self.username,
            password=self.password,
            path=self.path,
            scheme=self.scheme,
            client_name=self.client_name,
        )
        # if user sets dsn for dest node
        # use username and password from dsn settings
        if config.alias_dsn:
            for dsn_name, dsn_url in config.alias_dsn.items():
                dsn = parse_url(dsn_url)
                if dsn.host == ip and dsn.port == port:
                    print(
                        f"Connect {ip}:{port} via dns settings of {dsn_name}",
                        file=sys.stderr,
                    )
                    connection = self.create_connection(
                        dsn.host,
                        dsn.port,
                        dsn.db,
                        dsn.password,
                        dsn.path,
                        dsn.scheme,
                        dsn.username,
                    )
                    break

        connection.connect()
        return self.execute_by_connection(connection, *args, **kwargs)

    def render_response(self, response, command_name):
        "Parses a response from the Redis server"
        logger.info(f"[Redis-Server] Response: {response}")
        if config.raw:
            callback = OutputRender.render_raw
        # if in transaction, use queue render first
        elif config.transaction:
            callback = renders.OutputRender.render_transaction_queue
        else:
            callback = OutputRender.get_render(command_name=command_name)
        rendered = callback(response)
        logger.info(f"[render result] {rendered}")
        return rendered

    def monitor(self):
        """Redis' MONITOR command:
        https://redis.io/commands/monitor
        This command need to read from a stream resp, so
        it's different
        """
        while 1:
            response = self.connection.read_response()
            if config.raw:
                yield OutputRender.render_raw(response)
            else:
                yield OutputRender.render_bulk_string_decode(response)

    def subscribing(self):
        while 1:
            response = self.connection.read_response()
            if config.raw:
                yield OutputRender.render_raw(response)
            else:
                yield OutputRender.render_subscribe(response)

    def unsubscribing(self):
        "unsubscribe from all channels"
        response = self.execute("UNSUBSCRIBE")
        if config.raw:
            yield OutputRender.render_raw(response)
        else:
            yield OutputRender.render_subscribe(response)

    def split_command_and_pipeline(self, rawinput, completer: IRedisCompleter):
        """
        split user raw input to redis command and shell pipeline.
        eg:
        GET json | jq .key
        return: GET json, jq . key
        """
        grammar = completer.get_completer(input_text=rawinput).compiled_grammar
        matched = grammar.match(rawinput)
        if not matched:
            # invalid command!
            return rawinput, None
        variables = matched.variables()
        shell_command = variables.get("shellcommand")
        if shell_command:
            redis_command = rawinput.replace(shell_command, "")
            shell_command = shell_command.lstrip("| ")
            return redis_command, shell_command
        return rawinput, None

    def send_command(self, raw_command, completer=None):  # noqa
        """
        Send raw_command to redis-server, return parsed response.

        :param raw_command: text raw_command, not parsed
        :param completer: RedisGrammarCompleter will update completer
            based on redis response. eg: update key completer after ``keys``
            raw_command
        """
        if completer is None:  # not in a tty
            redis_command, shell_command = raw_command, None
        else:
            redis_command, shell_command = self.split_command_and_pipeline(
                raw_command, completer
            )
        logger.info(f"[Prepare command] Redis: {redis_command}, Shell: {shell_command}")
        try:
            try:
                command_name, args = split_command_args(redis_command)
            except (InvalidArguments, AmbiguousCommand):
                logger.warn(
                    "This is not a iredis known command, send to redis-server anyway..."
                )
                command_name, args = split_unknown_args(redis_command)

            logger.info(f"[Split command] command: {command_name}, args: {args}")
            input_command_upper = command_name.upper()
            # Confirm for dangerous command
            if config.warning:
                confirm = confirm_dangerous_command(input_command_upper)
                if confirm is True:
                    print("Your Call!!", file=sys.stderr)
                elif confirm is False:
                    print("Canceled!", file=sys.stderr)
                    return
                # None: continue...

            self.pre_hook(raw_command, command_name, args, completer)
            # if raw_command is not supposed to send to server
            if input_command_upper in CLIENT_COMMANDS:
                logger.info(f"{input_command_upper} is an iredis command.")
                yield from self.client_execute_command(command_name, *args)
                return

            redis_resp = self.execute(command_name, *args)
            # if shell_command and enable shell, do not render, just run in shell pipe and show the
            # subcommand's stdout/stderr
            if shell_command and config.shell:
                # pass the raw response of redis to shell command
                stdin = OutputRender.render_raw(redis_resp)
                run(shell_command, input=stdin, shell=True)
                return

            self.after_hook(raw_command, command_name, args, completer, redis_resp)
            yield self.render_response(redis_resp, command_name)

            # FIXME generator response do not support pipeline
            if input_command_upper == "MONITOR":
                # TODO special render for monitor
                try:
                    yield from self.monitor()
                except KeyboardInterrupt:
                    pass
            elif input_command_upper in [
                "SUBSCRIBE",
                "PSUBSCRIBE",
            ]:  # enter subscribe mode
                try:
                    yield from self.subscribing()
                except KeyboardInterrupt:
                    yield from self.unsubscribing()
        except Exception as e:
            logger.exception(e)
            if config.raw:
                render_callback = OutputRender.render_raw
            else:
                render_callback = OutputRender.render_error
            yield render_callback(f"ERROR {str(e)}".encode())
        finally:
            config.withscores = False

    def after_hook(self, command, command_name, args, completer, response):
        # === After hook ===
        # SELECT db on AUTH
        if command_name.upper() == "AUTH":
            if self.db:
                select_result = self.execute("SELECT", self.db)
                if nativestr(select_result) != "OK":
                    raise ConnectionError("Invalid Database")
            # When the connection is TimeoutError or ConnectionError, reconnect the connection will use it
            self.connection.password = args[0]
        elif command_name.upper() == "SELECT":
            logger.debug("[After hook] Command is SELECT, change self.db.")
            self.db = int(args[0])
            # When the connection is TimeoutError or ConnectionError, reconnect the connection will use it
            self.connection.db = self.db
        elif command_name.upper() == "MULTI":
            logger.debug("[After hook] Command is MULTI, start transaction.")
            config.transaction = True

        if completer:
            completer.update_completer_for_response(command_name, args, response)

    def pre_hook(self, command, command_name, args, completer: IRedisCompleter):
        """
        Before execute command, patch completers first.
        Eg: When user run `GET foo`, key completer need to
          touch foo.

        Only works when compile-grammar thread is done.
        """
        if command_name.upper() == "HELLO":
            raise NotSupport("IRedis currently not support RESP3, sorry about that.")
        # TRANSACTION state change
        if command_name.upper() in ["EXEC", "DISCARD"]:
            logger.debug(f"[After hook] Command is {command_name}, unset transaction.")
            config.transaction = False
        # score display for sorted set
        if command_name.upper() in ["ZSCAN", "ZPOPMAX", "ZPOPMIN"]:
            config.withscores = True

        # TODO should we using escape_decode on all strings??
        if command_name.upper() == "RESTORE":
            for i, a in enumerate(args):
                serialized_value = codecs.escape_decode(a)[0]
                args[i] = serialized_value

        # not a tty
        if not completer:
            logger.warning(
                "[Pre patch completer] Complter is None, not a tty, "
                "not patch completers, not set withscores"
            )
            return
        completer.update_completer_for_input(command)

        redis_grammar = completer.get_completer(command).compiled_grammar
        m = redis_grammar.match(command)
        if not m:
            # invalid command!
            return
        variables = m.variables()
        # zset withscores
        withscores = variables.get("withscores")
        if withscores:
            config.withscores = True

    def do_help(self, *args):
        command_docs_name = "-".join(args).lower()
        command_summary_name = " ".join(args).upper()
        try:
            doc = (
                files(commands_data)
                .joinpath(f"{command_docs_name}.md")
                .read_text(encoding="utf-8")
            )
        except FileNotFoundError:
            raise NotRedisCommand(
                f"{command_summary_name} is not a valid Redis command."
            )
        rendered_detail = markdown.render(doc)
        summary_dict = commands_summary[command_summary_name]

        available_version = summary_dict.get("since", "?")
        server_version = config.version
        # FIXME anything strange with single quotes?
        logger.debug(f"[--version--] '{server_version}'")
        try:
            is_available = version_parse(server_version) > version_parse(
                available_version
            )
        except Exception as e:
            logger.exception(e)
            is_available = None

        if is_available:
            available_text = f"(Available on your redis-server: {server_version})"
        elif is_available is False:
            available_text = f"(Not available on your redis-server: {server_version})"
        else:
            available_text = ""
        since_text = f"{available_version} {available_text}"

        summary = [
            ("", "\n"),
            ("class:doccommand", "  " + command_summary_name),
            ("", "\n"),
            ("class:dockey", "  summary: "),
            ("", summary_dict.get("summary", "No summary")),
            ("", "\n"),
            ("class:dockey", "  complexity: "),
            ("", summary_dict.get("complexity", "?")),
            ("", "\n"),
            ("class:dockey", "  since: "),
            ("", since_text),
            ("", "\n"),
            ("class:dockey", "  group: "),
            ("", summary_dict.get("group", "?")),
            ("", "\n"),
            ("class:dockey", "  syntax: "),
            ("", command_summary_name),  # command
            *compose_command_syntax(summary_dict, style_class=""),  # command args
            ("", "\n\n"),
        ]

        to_render = FormattedText(summary + rendered_detail)
        if config.raw:
            return convert_formatted_text_to_bytes(to_render)
        return to_render

    def do_peek(self, key):
        """
        PEEK command implementation.

        It's a generator, will run different redis commands based on the key's
        type, yields FormattedText once a command reached result.

        Redis current supported types:
            string, list, set, zset, hash and stream.
        """

        def _string(key):
            strlen = self.execute("strlen", key)
            yield FormattedText([("class:dockey", "strlen: "), ("", str(strlen))])

            value = self.execute("GET", key)
            yield FormattedText(
                [
                    ("class:dockey", "value: "),
                    ("", renders.OutputRender.render_bulk_string(value)),
                ]
            )

        def _list(key):
            llen = self.execute("llen", key)
            yield FormattedText([("class:dockey", "llen: "), ("", str(llen))])
            if llen <= 20:
                contents = self.execute(f"LRANGE {key} 0 -1")
            else:
                first_10 = self.execute(f"LRANGE {key} 0 9")
                last_10 = self.execute(f"LRANGE {key} -10 -1")
                contents = first_10 + [f"{llen-20} elements was omitted ..."] + last_10
            yield FormattedText([("class:dockey", "elements: ")])
            yield renders.OutputRender.render_list(contents)

        def _set(key):
            cardinality = self.execute("scard", key)
            yield FormattedText(
                [("class:dockey", "cardinality: "), ("", str(cardinality))]
            )
            if cardinality <= 20:
                contents = self.execute("smembers", key)
                yield FormattedText([("class:dockey", "members: ")])
                yield renders.OutputRender.render_list(contents)
            else:
                _, contents = self.execute(f"sscan {key} 0 count 20")
                first_n = len(contents)
                yield FormattedText([("class:dockey", f"members (first {first_n}): ")])
                yield renders.OutputRender.render_members(contents)
                # TODO update completers

        def _zset(key):
            count = self.execute(f"zcount {key} -inf +inf")
            yield FormattedText([("class:dockey", "zcount: "), ("", str(count))])
            if count <= 20:
                contents = self.execute(f"zrange {key} 0 -1 withscores")
                yield FormattedText([("class:dockey", "members: ")])
                yield renders.OutputRender.render_members(contents)
            else:
                _, contents = self.execute(f"zscan {key} 0 count 20")
                first_n = len(contents) // 2
                yield FormattedText([("class:dockey", f"members (first {first_n}): ")])
                config.withscores = True
                output = renders.OutputRender.render_members(contents)
                config.withscores = False
                yield output

        def _hash(key):
            hlen = self.execute(f"hlen {key}")
            yield FormattedText([("class:dockey", "hlen: "), ("", str(hlen))])
            if hlen <= 20:
                contents = self.execute(f"hgetall {key}")
                yield FormattedText([("class:dockey", "fields: ")])
            else:
                _, contents = self.execute(f"hscan {key} 0 count 20")
                first_n = len(contents) // 2
                yield FormattedText([("class:dockey", f"fields (first {first_n}): ")])
            yield renders.OutputRender.render_hash_pairs(contents)

        def _stream(key):
            xinfo = self.execute("xinfo stream", key)
            yield FormattedText([("class:dockey", "XINFO: ")])
            yield renders.OutputRender.render_list(xinfo)

        # in case the result is too long, we yield only once so the outputer
        # can pager it.
        peek_response = []
        key_type = nativestr(self.execute("type", key))
        if key_type == "none":
            yield f"{key} doesn't exist."
            return

        encoding = nativestr(self.execute("object encoding", key))

        # use `memory usage` to get memory, this command available from redis4.0
        mem = ""
        if config.version and version_parse(config.version) >= version_parse("4.0.0"):
            memory_usage_value = str(self.execute("memory usage", key))
            mem = f"  mem: {memory_usage_value} bytes"

        ttl = str(self.execute("ttl", key))

        key_info = f"{key_type} ({encoding}){mem}, ttl: {ttl}"

        # FIXME raw write_result parse FormattedText
        peek_response.append(FormattedText([("class:dockey", "key: "), ("", key_info)]))

        detail_action_fun = {
            "string": _string,
            "list": _list,
            "set": _set,
            "zset": _zset,
            "hash": _hash,
            "stream": _stream,
        }[key_type]
        detail = list(detail_action_fun(key))
        peek_response.extend(detail)

        # merge them into only one FormattedText
        flat_formatted_text_pair = []
        for index, formatted_text in enumerate(peek_response):
            for ft in formatted_text:
                flat_formatted_text_pair.append(ft)
            if index < len(peek_response) - 1:
                flat_formatted_text_pair.append(renders.NEWLINE_TUPLE)

        if config.raw:
            yield convert_formatted_text_to_bytes(flat_formatted_text_pair)
            return
        yield FormattedText(flat_formatted_text_pair)



import re
import csv
import json
import logging
import functools
from importlib.resources import files

from .utils import timer, strip_quote_args
from .exceptions import InvalidArguments, AmbiguousCommand
from . import data as project_data

logger = logging.getLogger(__name__)


def _load_command_summary():
    commands_summary = json.loads(
        files(project_data).joinpath("commands.json").read_text(encoding="utf-8")
    )
    return commands_summary


def _load_command():
    """
    load command information from file.
    :returns:
        - original_commans: dict, command name : Command
        - command_group: dict, group_name: command_names
    """
    first_line = True
    command2callback = {}
    command2syntax = {}
    groups = {}
    with (
        files(project_data)
        .joinpath("command_syntax.csv")
        .open("r", encoding="utf-8") as command_syntax
    ):
        csvreader = csv.reader(command_syntax)
        for line in csvreader:
            if first_line:
                first_line = False
                continue
            group, command, syntax, func_name = line
            command2callback[command] = func_name
            command2syntax[command] = syntax
            groups.setdefault(group, []).append(command)

    return command2callback, command2syntax, groups


def _load_dangerous():
    """
    Load dangerous commands from csv file.
    """
    first_line = True
    dangerous_command = {}
    with (
        files(project_data)
        .joinpath("dangerous_commands.csv")
        .open("r", encoding="utf-8") as dangerous_file
    ):
        csvreader = csv.reader(dangerous_file)
        for line in csvreader:
            if first_line:
                first_line = False
                continue
            command, reason = line
            dangerous_command[command] = reason
    return dangerous_command


timer("[Loader] Start loading commands file...")
command2callback, command2syntax, groups = _load_command()
# all redis command strings, in UPPER case
# NOTE: Must sort by length, to match longest command first
all_commands = sorted(
    list(command2callback.keys()) + ["HELP"], key=lambda x: len(x), reverse=True
)
# load commands information from redis-doc/commands.json
commands_summary = _load_command_summary()
# add iredis' commands' summary
commands_summary.update(
    {
        "HELP": {
            "summary": "Show documents for a Redis command.",
            "complexity": "O(1).",
            "arguments": [{"name": "command", "type": "string"}],
            "since": "1.0",
            "group": "iredis",
        },
        "CLEAR": {
            "summary": "Clear the screen like bash clear.",
            "complexity": "O(1).",
            "since": "1.0",
            "group": "iredis",
        },
        "EXIT": {
            "summary": "Exit iredis.",
            "complexity": "O(1).",
            "since": "1.0",
            "group": "iredis",
        },
        "PEEK": {
            "summary": "Get the key's type and value.",
            "arguments": [{"name": "key", "type": "key"}],
            "complexity": "O(1).",
            "since": "1.0",
            "group": "iredis",
        },
    }
)
timer("[Loader] Finished loading commands.")
dangerous_commands = _load_dangerous()


@functools.lru_cache(maxsize=2048)
def split_command_args(command):
    """
    Split Redis command text into command and args.

    :param command: redis command string, with args
    """
    command = command.strip()
    for command_name in all_commands:
        # for command that is partially input, like `command in`, we should
        # match with `command info`, otherwise, `command in` will result in
        # `command` with `args` is ('in') which is an invalid case.
        normalized_input_command = " ".join(command.split()).upper()
        if (
            re.search(r"\s", command)
            and command_name.startswith(normalized_input_command)
            and command_name != normalized_input_command
        ):
            raise AmbiguousCommand("command is not finished")
        # allow multiple space in user input command
        command_allow_multi_spaces = "[ ]+".join(command_name.split())
        matcher = re.match(rf"({command_allow_multi_spaces})( |$)", command.upper())
        if matcher:
            matched_command_len = len(matcher.group(1))
            input_command = command[:matched_command_len]
            input_args = command[matcher.end() :]
            break
    else:
        raise InvalidArguments(f"`{command}` is not a valid Redis Command")

    args = list(strip_quote_args(input_args))

    return input_command, args


def split_unknown_args(command):
    """
    Split user's input into command and args.
    """
    command = command.strip()
    input_command, *input_args = command.split(" ")
    return input_command, list(strip_quote_args(" ".join(input_args)))



"""
Markdown render.

use https://github.com/lepture/mistune render to html, then print with my style.
"""

import logging
import mistune
import re
from prompt_toolkit.formatted_text import to_formatted_text, HTML

logger = logging.getLogger(__name__)


class TerminalRender(mistune.HTMLRenderer):
    def _to_title(self, text):
        return f"{text}\n{'='*len(text)}\n"

    def paragraph(self, text):
        return text + "\n\n"

    def block_code(self, code, info=None):
        code = "\n".join(["  " + line for line in code.splitlines()])
        return super().block_code(code)

    def heading(self, text, level):
        if level == 2:
            header_text = self._to_title(text)
            return super().heading(header_text, 2)
        return super().heading(self._to_title(text), level)

    def list(self, body, ordered, *args, **kwargs):
        """Rendering list tags like ``<ul>`` and ``<ol>``.

        :param body: body contents of the list.
        :param ordered: whether this list is ordered or not.
        """
        tag = "ul"
        if ordered:
            tag = "ol"
        return "<{}>{}</{}>\n".format(tag, body, tag)

    def list_item(self, text, *args):
        """Rendering list item snippet. Like ``<li>``."""
        return "<li> * %s</li>\n" % text


renderer = TerminalRender()
markdown_render = mistune.Markdown(renderer)

# replace redis doc's title (and following newlines & spaces)
# with markdown's second level title
redisdoc_title_re = re.compile(r"^@(\w+) *(?:\n+|$)")


def replace_to_markdown_title(original):
    replaced = redisdoc_title_re.sub(r"## \g<1>", original)
    return replaced


def render(text):
    replaced = replace_to_markdown_title(text)
    html_text = markdown_render(replaced)
    logger.debug(f"[Document] {html_text} ..."[:20])
    return to_formatted_text(HTML(html_text))



import os
import logging
import sys
import time
from pathlib import Path
import platform

import click
from prompt_toolkit import PromptSession
from prompt_toolkit.history import FileHistory
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit import print_formatted_text
from prompt_toolkit.formatted_text import FormattedText
from prompt_toolkit.key_binding.bindings.named_commands import (
    register as prompt_register,
)

from .client import Client
from .key_bindings import kb as key_bindings
from .style import STYLE
from .config import config, load_config_files
from .processors import UserInputCommand, UpdateBottomProcessor, PasswordProcessor
from .bottom import BottomToolbar
from .utils import timer, exit, convert_formatted_text_to_bytes, parse_url
from .completers import IRedisCompleter
from .lexer import IRedisLexer
from . import __version__

logger = logging.getLogger(__name__)


class SkipAuthFileHistory(FileHistory):
    """Exactly like FileHistory, but won't save `AUTH` command into history
    file."""

    def append_string(self, string: str) -> None:
        if string.lstrip().upper().startswith("AUTH"):
            return
        super().append_string(string)


def setup_log():
    if config.log_location:
        logging.basicConfig(
            filename=os.path.expanduser(config.log_location),
            filemode="a",
            format="%(levelname)5s %(message)s",
            level="DEBUG",
        )
    else:
        logging.disable(logging.CRITICAL)
    logger.info("------ iRedis ------")


def greetings():
    iredis_version = f"iredis  {__version__} (Python {platform.python_version()})"
    if config.no_version_reason:
        reason = f"({config.no_version_reason})"
    else:
        reason = ""

    server_version = f"redis-server  {config.version} {reason}"
    home_page = "Home:   https://iredis.xbin.io/"
    issues = "Issues: https://github.com/laixintao/iredis/issues"
    display = "\n".join([iredis_version, server_version, home_page, issues])
    if config.raw:
        display = display.encode()
    write_result(display)


def print_help_msg(command):
    with click.Context(command) as ctx:
        click.echo(command.get_help(ctx))


def is_too_tall(text, max_height):
    if isinstance(text, FormattedText):
        text = convert_formatted_text_to_bytes(text)
    lines = len(text.split(b"\n"))
    return lines > max_height


def write_result(text, max_height=None):
    """
    When config.raw set to True, write text(must be bytes in that case)
    directly to stdout, same if text is bytes.

    :param text: is_raw: bytes or str, not raw: FormattedText
    :is_raw: bool
    """
    logger.info(f"Print result {type(text)}: {text}"[:200])

    # this function only handle bytes or FormattedText
    # if it's str, convert to bytes
    if isinstance(text, str):
        if config.decode:
            text = text.encode(config.decode)
        else:
            text = text.encode()

    # using pager if too tall
    if max_height and config.enable_pager and is_too_tall(text, max_height):
        if isinstance(text, FormattedText):
            text = convert_formatted_text_to_bytes(text)
            os.environ["LESS"] = "-SRX"
        # click.echo_via_pager only accepts str
        if config.decode:
            text = text.decode(config.decode)
        else:
            text = text.decode()
        # TODO current pager doesn't support colors
        click.echo_via_pager(text)
        return

    if isinstance(text, bytes):
        sys.stdout.buffer.write(text)
        sys.stdout.write("\n")
    else:
        print_formatted_text(text, end="", style=STYLE)
        print_formatted_text()


class Rainbow:
    color = [
        "#cc2244",
        "#bb4444",
        "#996644",
        "#cc8844",
        "#ccaa44",
        "#bbaa44",
        "#99aa44",
        "#778844",
        "#55aa44",
        "#33aa44",
        "#11aa44",
        "#11aa66",
        "#11aa88",
        "#11aaaa",
        "#11aacc",
        "#11aaee",
    ]

    def __init__(self):
        self.current = -1
        self.forward = 1

    def __iter__(self):
        return self

    def __next__(self):
        self.current += self.forward
        if 0 <= self.current < len(self.color):
            # not to the end
            return self.color[self.current]
        else:
            self.forward = -self.forward
            self.current += 2 * self.forward
            return self.color[self.current]


def prompt_message(client):
    text = str(client)
    if config.rainbow:
        return list(zip(Rainbow(), text))
    return text


def repl(client, session, start_time):
    command_holder = UserInputCommand()
    timer(f"First REPL command enter, time cost: {time.time() - start_time}")

    while True:
        logger.info("↓↓↓↓" * 10)
        logger.info("REPL waiting for command...")

        try:
            command = session.prompt(
                prompt_message(client),
                bottom_toolbar=(
                    BottomToolbar(command_holder).render if config.bottom_bar else None
                ),
                input_processors=[
                    UpdateBottomProcessor(command_holder, session),
                    PasswordProcessor(),
                ],
                rprompt=lambda: "<transaction>" if config.transaction else None,
                key_bindings=key_bindings,
                enable_suspend=True,
            )

        except KeyboardInterrupt:
            logger.warning("KeyboardInterrupt!")
            continue
        except EOFError:
            exit()
        command = command.strip()
        logger.info(f"[Command] {command}")

        # blank input
        if not command:
            continue

        try:
            answers = client.send_command(command, session.completer)
            for answer in answers:
                write_result(
                    answer,
                    # -1 is because 127.0.0.1:6379> takes one line
                    session.output.get_size().rows - session.reserve_space_for_menu - 1,
                )
        # Error with previous command or exception
        except Exception as e:
            logger.exception(e)
            # TODO red error color
            print("(error)", str(e))


RAW_HELP = """
Use raw formatting for replies (default when STDOUT is not a tty). \
However, you can use --no-raw to force formatted output even \
when STDOUT is not a tty.
"""
DECODE_HELP = """
decode response, default is No decode, which will output all bytes literals.
"""
RAINBOW = "Display colorful prompt."
DSN_HELP = """
Use DSN configured into the [alias_dsn] section of iredisrc file. \
(Can set with env `IREDIS_DSN`)
"""
URL_HELP = """
Use Redis URL to indicate connection(Can set with env `IREDIS_URL`), Example:
    redis://[[username]:[password]]@localhost:6379/0
    rediss://[[username]:[password]]@localhost:6379/0
    unix://[[username]:[password]]@/path/to/socket.sock?db=0
"""
SHELL = """Allow to run shell commands, default to True."""
PAGER_HELP = """Using pager when output is too tall for your window, default to True."""
VERIFY_SSL_HELP = """Set the TLS certificate verification strategy"""


# command line entry here...
@click.command()
@click.pass_context
@click.option("-h", help="Server hostname (default: 127.0.0.1).", default="127.0.0.1")
@click.option("-p", help="Server port (default: 6379).", default="6379")
@click.option(
    "-s", "--socket", default=None, help="Server socket (overrides hostname and port)."
)
@click.option("-n", help="Database number.(overwrites dsn/url's db number)", default=0)
@click.option(
    "-u",
    "--username",
    help="User name used to auth, will be ignore for redis version < 6.",
)
@click.option("-a", "--password", help="Password to use when connecting to the server.")
@click.option("--url", default=None, envvar="IREDIS_URL", help=URL_HELP)
@click.option("-d", "--dsn", default=None, envvar="IREDIS_DSN", help=DSN_HELP)
@click.option(
    "--newbie/--no-newbie",
    default=None,
    is_flag=True,
    help="Show command hints and useful helps.",
)
@click.option(
    "--iredisrc",
    default="~/.iredisrc",
    envvar="IREDIS_CONFIG",
    help=(
        "Config file for iredis, default is ~/.iredisrc. "
        "You can also set config path via environment variable `IREDIS_CONFIG`."
    ),
)
@click.option("--decode", default=None, help=DECODE_HELP)
@click.option("--client_name", help="Assign a name to the current connection.")
@click.option("--raw/--no-raw", default=None, is_flag=True, help=RAW_HELP)
@click.option("--rainbow/--no-rainbow", default=None, is_flag=True, help=RAINBOW)
@click.option("--shell/--no-shell", default=None, is_flag=True, help=SHELL)
@click.option("--pager/--no-pager", default=None, is_flag=True, help=PAGER_HELP)
@click.option(
    "--greetings/--no-greetings",
    default=None,
    is_flag=True,
    help="Enable or disable greeting messages",
)
@click.option(
    "--verify-ssl",
    default=None,
    type=click.Choice(["none", "optional", "required"]),
    help=VERIFY_SSL_HELP,
)
@click.option(
    "--prompt",
    default=None,
    help=(
        "Prompt format (supported interpolations: {client_name}, {db}, {host}, {path},"
        " {port}, {username}, {client_addr}, {client_id})."
    ),
)
@click.option(
    "--natmap",
    default=None,
    help=(
        "NAT map for Redis cluster behind SSH tunnels. "
        "Format: remoteHost:remotePort:localHost:localPort "
        "(comma-separated for multiple nodes)."
    ),
)
@click.version_option()
@click.argument("cmd", nargs=-1)
def gather_args(
    ctx,
    h,
    p,
    n,
    username,
    password,
    client_name,
    newbie,
    iredisrc,
    decode,
    raw,
    rainbow,
    cmd,
    dsn,
    url,
    socket,
    shell,
    pager,
    greetings,
    verify_ssl,
    prompt,
    natmap,
):
    """
    IRedis: Interactive Redis

    When no command is given, IRedis starts in interactive mode.

    \b
    Examples:
      - iredis
      - iredis -d dsn
      - iredis -h 127.0.0.1 -p 6379
      - iredis -h 127.0.0.1 -p 6379 -a <password>
      - iredis --url redis://localhost:7890/3

    Type "help" in interactive mode for information on available commands
    and settings.
    """
    load_config_files(iredisrc)
    setup_log()
    logger.info(
        f"[commandline args] host={h}, port={p}, db={n}, user={username},"
        f" newbie={newbie}, iredisrc={iredisrc}, decode={decode}, raw={raw}, cmd={cmd},"
        f" rainbow={rainbow}."
    )
    # raw config
    if raw is not None:
        config.raw = raw
    if not sys.stdout.isatty():
        config.raw = True

    if newbie is not None:
        config.newbie_mode = newbie

    if decode is not None:
        config.decode = decode
    if rainbow is not None:
        config.rainbow = rainbow
    if shell is not None:
        config.shell = shell
    if pager is not None:
        config.enable_pager = pager
    if verify_ssl is not None:
        config.verify_ssl = verify_ssl
    if greetings is not None:
        config.greetings = greetings

    if natmap is not None:
        for entry in natmap.split(","):
            parts = entry.strip().split(":")
            remote_host, remote_port, local_host, local_port = parts
            config.natmap[f"{remote_host}:{remote_port}"] = (
                local_host,
                int(local_port),
            )

    return ctx


@prompt_register("edit-and-execute-command")
def edit_and_execute(event):
    """Different from the prompt-toolkit default, we want to have a choice not
    to execute a query after editing, hence validate_and_handle=False."""
    buff = event.current_buffer
    # this will prevent running command immediately when exit editor.
    buff.open_in_editor(validate_and_handle=False)


def resolve_dsn(dsn):
    try:
        dsn_uri = config.alias_dsn[dsn]
    except KeyError:
        click.secho(
            "Could not find the specified DSN in the config file. "
            'Please check the "[alias_dsn]" section in your '
            "iredisrc.",
            err=True,
            fg="red",
        )
        sys.exit(1)
    return dsn_uri


def create_client(params):
    """
    Create a Client.
    :param params: commandline params.
    """
    host = params["h"]
    port = params["p"]
    db = params["n"]
    username = params["username"]
    password = params["password"]
    client_name = params["client_name"]
    prompt = params["prompt"]
    verify_ssl = params["verify_ssl"]

    dsn_from_url = None
    dsn = params["dsn"]
    if config.alias_dsn and dsn:
        dsn_uri = resolve_dsn(dsn)
        dsn_from_url = parse_url(dsn_uri)
    if params["url"]:
        dsn_from_url = parse_url(params["url"])
    if dsn_from_url:
        # db from command lint options should be high priority
        db = db if db else dsn_from_url.db
        verify_ssl = verify_ssl or dsn_from_url.verify_ssl
        return Client(
            host=dsn_from_url.host,
            port=dsn_from_url.port,
            db=db,
            password=dsn_from_url.password,
            path=dsn_from_url.path,
            scheme=dsn_from_url.scheme,
            username=dsn_from_url.username,
            client_name=client_name,
            prompt=prompt,
            verify_ssl=verify_ssl,
        )
    if params["socket"]:
        return Client(
            scheme="unix",
            path=params["socket"],
            db=db,
            username=username,
            password=password,
            client_name=client_name,
            prompt=prompt,
        )
    return Client(
        host=host,
        port=port,
        db=db,
        username=username,
        password=password,
        client_name=client_name,
        prompt=prompt,
        verify_ssl=verify_ssl,
    )


def main():
    enter_main_time = time.time()  # just for logs

    # invoke in non-standalone mode to gather args
    ctx = None
    try:
        ctx = gather_args.main(standalone_mode=False)
    except click.exceptions.NoSuchOption as nosuchoption:
        nosuchoption.show()
    except click.exceptions.BadOptionUsage as badoption:
        if badoption.option_name == "-h":
            # -h without host, is short command for --help
            # like redis-cli
            print_help_msg(gather_args)
        return
    if not ctx:  # called help
        return

    # redis client
    client = create_client(ctx.params)

    if not sys.stdin.isatty():
        for line in sys.stdin.readlines():
            logger.debug(f"[Command stdin] {line}")
            for answer in client.send_command(line, None):
                write_result(answer)
        return

    # no interactive mode, directly run a command
    if ctx.params["cmd"]:
        answers = client.send_command(" ".join(ctx.params["cmd"]), None)
        for answer in answers:
            write_result(answer)
        logger.warning("[OVER] command executed, exit...")
        return

    # prompt session
    session = PromptSession(
        history=SkipAuthFileHistory(Path(os.path.expanduser(config.history_location))),
        style=STYLE,
        auto_suggest=AutoSuggestFromHistory(),
        complete_while_typing=True,
        lexer=IRedisLexer(),
        completer=IRedisCompleter(
            hint=config.newbie_mode, completion_casing=config.completion_casing
        ),
        enable_open_in_editor=True,
        tempfile_suffix=".redis",
    )

    # print hello message
    if config.greetings:
        greetings()
    repl(client, session, enter_main_time)



"""
Render redis-server responses.
This module will be auto loaded to callbacks.

func(redis-response) -> formatted result(str)
"""

import logging
import time
from packaging.version import parse as version_parse

from prompt_toolkit.formatted_text import FormattedText

from .commands import command2callback
from .config import config
from .utils import double_quotes, ensure_str, nativestr

logger = logging.getLogger(__name__)
NEWLINE_TUPLE = ("", "\n")
NIL_TUPLE = ("class:type", "(nil)")
NIL = FormattedText([NIL_TUPLE])
EMPTY_LIST = FormattedText([("class:type", "(empty list or set)")])


class OutputRender:
    """Render redis output"""

    @staticmethod
    def get_render(command_name):
        """Dynamic render output due to command name."""
        command_upper = " ".join(command_name.split()).upper()
        callback_name = command2callback.get(command_upper)

        # using `render_list_or_string` as default render.
        if callback_name is None:
            callback = OutputRender.render_list_or_string
        else:
            callback = getattr(
                OutputRender, callback_name, OutputRender.render_list_or_string
            )

        logger.info(
            f"[render] Find callback {callback_name}, for command: {command_name}"
        )
        return callback

    @staticmethod
    def render_raw(value):
        """
        Render for all kinds, list, string, bulkstring, int

        :return : bytes
        """
        if value is None:
            return b""
        if isinstance(value, bytes):
            return value
        if isinstance(value, int):
            return str(value).encode()
        if isinstance(value, str):
            return value.encode()
        if isinstance(value, list):
            return _render_raw_list(value)

    @staticmethod
    def render_bulk_string(value):
        if value is None:
            return NIL
        return double_quotes(ensure_str(value))

    @staticmethod
    def render_bulk_string_decode(value):
        """
        Only for server group commands, no double quoted, always displayed as
        utf-8 decoded.
        """
        decoded = nativestr(value)
        split = "\n".join(decoded.splitlines())  # get rid of last newline
        return split

    @staticmethod
    def render_nested_pair(value):
        """
        For redis internal responses.
        Always decode with utf-8
        Render nested list.
        Items come as pairs.
        """
        return FormattedText(_render_pair(value, 0))

    @staticmethod
    def render_int(value):
        if value is None:
            return NIL
        return FormattedText([("class:type", "(integer) "), ("", str(value))])

    @staticmethod
    def render_unixtime(value):
        rendered_int = OutputRender.render_int(value)
        explained_date = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(int(value)))
        rendered_int.extend(
            [
                NEWLINE_TUPLE,
                ("class:type", "(local time)"),
                ("", " "),
                ("", explained_date),
            ]
        )
        return rendered_int

    @staticmethod
    def render_time(value):
        unix_timestamp, millisecond = value[0].decode(), value[1].decode()
        explained_date = time.strftime(
            "%Y-%m-%d %H:%M:%S", time.localtime(int(unix_timestamp))
        )
        rendered = [
            ("class:type", "(unix timestamp) "),
            ("", unix_timestamp),
            NEWLINE_TUPLE,
            ("class:type", "(millisecond) "),
            ("", millisecond),
            NEWLINE_TUPLE,
            ("class:type", "(convert to local timezone) "),
            ("", f"{explained_date}.{millisecond}"),
        ]
        return FormattedText(rendered)

    @staticmethod
    def render_list(text, style="class:string"):
        """
        Render callback for redis Array Reply
        Note: Cloud be null in it.
        """
        str_items = []
        for item in text:
            if item is None:
                str_items.append(None)
            else:
                str_item = ensure_str(item)
                double_quoted = double_quotes(str_item)
                str_items.append(double_quoted)
        rendered = _render_list(text, str_items, style)
        return FormattedText(rendered)

    @staticmethod
    def render_list_or_string(text):
        if isinstance(text, list):
            return OutputRender.render_list(text)
        return OutputRender.render_bulk_string(text)

    @staticmethod
    def render_list_or_int_or_string(text):
        if isinstance(text, list):
            return OutputRender.render_list(text)
        if isinstance(text, int):
            return OutputRender.render_int(text)
        return OutputRender.render_bulk_string(text)

    @staticmethod
    def render_string_or_int(text):
        if isinstance(text, int):
            return OutputRender.render_int(text)
        return OutputRender.render_bulk_string(text)

    @staticmethod
    def render_error(error_msg):
        text = ensure_str(error_msg)
        return FormattedText([("class:type", "(error) "), ("class:error", text)])

    @staticmethod
    def render_simple_string(text):
        """
        If response is b'OK', render simple string always with success color.
        If Error happens, error will be rendered by ``render_error``
        """
        if text is None:
            return NIL
        text = ensure_str(text)
        return FormattedText([("class:success", text)])

    @staticmethod
    def render_help(raw):
        """
        render help text message.
        the command like ``ACL HELP`` and ``MEMORY HELP``
        will return a list of strings.
        we render it as plain text
        """
        return FormattedText([("class:string", _render_raw_list(raw).decode())])

    @staticmethod
    def render_transaction_queue(text):
        """
        Used when client session is in a transaction.

        Response message should be "QUEUE" or Error.
        """
        text = ensure_str(text)
        return FormattedText([("class:queued", text)])

    @staticmethod
    def render_members(items):
        if not config.withscores:
            return OutputRender.render_list(items, "class:member")

        if not items:
            return EMPTY_LIST
        str_items = ensure_str(items)

        members = [item for item in str_items[::2]]
        scores = [item for item in str_items[1::2]]
        logger.debug(f"[MEMBERS] {members}")
        logger.debug(f"[SCORES] {scores}")
        # render display
        double_quoted = double_quotes(members)
        index_width = len(str(len(double_quoted)))
        score_width = max(len(score) for score in scores)
        rendered = []
        for index, item in enumerate(double_quoted):
            index_const_width = f"{index+1:{index_width}})"
            rendered.append(("", index_const_width))
            # add a space between index and member
            rendered.append(("", " "))
            # add score
            rendered.append(("class:integer", f"{scores[index]:{score_width}} "))
            # add member
            if item is None:
                rendered.append(NIL_TUPLE)
            else:
                rendered.append(("class:member", item))

            # add a newline for eachline
            if index + 1 < len(double_quoted):
                rendered.append(NEWLINE_TUPLE)
        return FormattedText(rendered)

    @staticmethod
    def render_hash_pairs(response):
        # render hash pairs
        if not response:
            return EMPTY_LIST
        str_items = ensure_str(response)
        fields = str_items[0::2]
        values = str_items[1::2]
        # render display
        index_width = len(str(len(fields)))
        values_quoted = double_quotes(values)
        fields_quoted = double_quotes(fields)
        rendered = []
        for index, item in enumerate(fields_quoted):
            index_const_width = f"{index+1:{index_width}})"
            rendered.append(("", index_const_width))
            rendered.append(("", " "))
            rendered.append(("class:field", item))
            rendered.append(NEWLINE_TUPLE)
            rendered.append(("", " " * (len(index_const_width) + 1)))
            value = values_quoted[index]
            if value is None:
                rendered.append(NIL_TUPLE)
            else:
                rendered.append(("class:string", value))

            # add a newline for eachline
            if index + 1 < len(fields):
                rendered.append(NEWLINE_TUPLE)
        return FormattedText(rendered)

    @staticmethod
    def render_slowlog(raw):
        fields = ["Slow log id", "Start at", "Running time(μs)", "Command"]
        if version_parse(config.version) > version_parse("4.0"):
            fields.extend(["Client IP and port", "Client name"])

        rendered = []
        text = ensure_str(raw)
        index_width = len(str(len(text)))
        for index, slowlog in enumerate(text):
            index_str = f"{index+1:{index_width}}) "
            rendered.append(("", index_str))
            for field, value in zip(fields, slowlog):
                if field == "Command":
                    value = " ".join(value)
                if field != "Slow log id":
                    display_field = " " * len(index_str) + field
                else:
                    display_field = field
                logger.debug(f"field: {field}, value: {value}")
                rendered.extend(
                    [
                        ("class:field", f"{display_field}: "),
                        ("class:string", value),
                        NEWLINE_TUPLE,
                    ]
                )

        return FormattedText(rendered[:-1])

    @staticmethod
    def render_subscribe(raw):
        """
        message type;
        channel;
        message;
        see: https://redis.io/topics/pubsub#format-of-pushed-messages
        """
        logger.info(raw)
        if raw[1] is None:
            raw[1] = "all"
        mtype, *channel, message = ensure_str(raw)
        # PUNSUBSCRIBE, 4 args
        channel = ":".join(channel)
        return FormattedText(
            [
                ("", f"{mtype:<9} from "),  # 9 is len("subscribe")
                ("class:channel", channel),
                ("", ": "),  # 9 is len("subscribe")
                ("class:string", f"{message}"),
            ]
        )

    @staticmethod
    def command_keys(items):
        return OutputRender.render_list(items, "class:key")

    @staticmethod
    def command_scan(response):
        """
        Render Scan command result.
        see: https://redis.io/commands/scan
        """
        return _render_scan(OutputRender.command_keys, response)

    @staticmethod
    def command_sscan(response):
        return _render_scan(OutputRender.render_members, response)

    @staticmethod
    def command_zscan(response):
        return _render_scan(OutputRender.render_members, response)

    @staticmethod
    def command_hscan(response):
        return _render_scan(OutputRender.render_hash_pairs, response)

    @staticmethod
    def command_hkeys(response):
        return OutputRender.render_list(response, "class:field")

    @staticmethod
    def render_bytes(response):
        return response.rstrip(b"\n")  # there is a new line in `write_result`

    @staticmethod
    def default_render(text):
        pass


def _render_raw_list(bytes_items):
    flatten_items = []
    for item in bytes_items:
        if item is None:
            flatten_items.append(b"")
        elif isinstance(item, bytes):
            flatten_items.append(item)
        elif isinstance(item, int):
            flatten_items.append(str(item).encode())
        elif isinstance(item, str):
            flatten_items.append(item.encode())
        elif isinstance(item, list):
            flatten_items.append(_render_raw_list(item))
    return b"\n".join(flatten_items)


def _render_list(byte_items, str_items, style=None, pre_space=0):
    """Complute the newline/number-width/lineno,
    render list to FormattedText
    """
    if not str_items:
        return EMPTY_LIST

    index_width = len(str(len(str_items)))
    rendered = []
    for index, item in enumerate(str_items):
        indent_spaces = (index + 1 != 1) * pre_space * " "
        if indent_spaces:
            rendered.append(("", indent_spaces))  # add a space before item

        index_const_width = f"{index+1:{index_width}})"
        rendered.append(("", index_const_width))
        # list item
        rendered.append(("", " "))  # add a space before item
        if item is None:
            rendered.append(NIL_TUPLE)
        elif isinstance(item, str):
            rendered.append((style, item))
        else:  # it's a nested list
            # if config.raw == True, never will get there
            sublist = _render_list(None, item, style, pre_space + index_width + 2)
            rendered.extend(sublist)

        # add a newline for eachline
        if index + 1 < len(str_items):
            rendered.append(NEWLINE_TUPLE)
    return rendered


def _render_scan(render_response, response):
    cursor, responses = response

    rendered = [
        ("class:type", "(cursor) "),
        ("class:integer", cursor if isinstance(cursor, str) else cursor.decode()),
        ("", "\n"),
    ]
    rendered_keys = render_response(responses)
    return FormattedText(rendered + rendered_keys)


def _render_pair(pairs, indent):
    keys = [item for item in pairs[::2]]
    values = [item for item in pairs[1::2]]
    rendered = []
    for key, value in zip(keys, values):
        key = ensure_str(key, decode="utf-8")
        value = ensure_str(value, decode="utf-8")
        rendered.append(("class:string", f"{' '*4*indent}{key}: "))
        if isinstance(value, list):
            rendered.append(NEWLINE_TUPLE)
            rendered.extend(_render_pair(value, indent + 1))
        else:
            rendered.append(("class:value", value))
        rendered.append(NEWLINE_TUPLE)
    return rendered[:-1]  # remove last newline


# TODO
# special list render, bzpopmax, key-value pair



import re
import sys
import time
import logging
from collections import namedtuple
from urllib.parse import parse_qs, unquote, urlparse

from prompt_toolkit.formatted_text import FormattedText

from iredis.exceptions import InvalidArguments

logger = logging.getLogger(__name__)

_last_timer = time.time()
_timer_counter = 0
separator = re.compile(r"\s")
logger.debug(f"[timer] start on {_last_timer}")


def timer(title):
    global _last_timer
    global _timer_counter

    now = time.time()
    tick = now - _last_timer
    logger.debug(f"[timer{_timer_counter:2}] {tick:.8f} -> {title}")

    _last_timer = now
    _timer_counter += 1


def nativestr(x):
    return x if isinstance(x, str) else x.decode("utf-8", "replace")


def literal_bytes(b):
    if isinstance(b, bytes):
        return str(b)[2:-1]
    return b


def nappend(word, c, pre_back_slash):
    if pre_back_slash and c == "n":  # \n
        word[-1] = "\n"
    else:
        word.append(c)


def strip_quote_args(s):
    """
    Given string s, split it into args.(Like bash paring)
    Handle with all quote cases.

    Raise ``InvalidArguments`` if quotes not match

    :return: args list.
    """
    word = []
    in_quote = None
    pre_back_slash = False
    for char in s:
        if in_quote:
            # close quote
            if char == in_quote:
                if not pre_back_slash:
                    yield "".join(word)
                    word = []
                    in_quote = None
                else:
                    # previous char is \ , merge with current "
                    word[-1] = char
            else:
                nappend(word, char, pre_back_slash)
        # not in quote
        else:
            # separator
            if separator.match(char):
                if word:
                    yield "".join(word)
                    word = []
            # open quotes
            elif char in ["'", '"']:
                in_quote = char
            else:
                nappend(word, char, pre_back_slash)
        if char == "\\" and not pre_back_slash:
            pre_back_slash = True
        else:
            pre_back_slash = False

    if word:
        yield "".join(word)
    # quote not close
    if in_quote:
        raise InvalidArguments("Invalid argument(s)")


type_convert = {"posix time": "time"}


def parse_argument_to_formatted_text(
    name, _type, is_option, token=None, style_class="bottom-toolbar"
):
    result = []
    if isinstance(name, str):
        _type = type_convert.get(_type, _type)
        if token:
            # Format: [TOKEN name] or TOKEN name
            if is_option:
                result.append((f"class:{style_class}.{_type}", f" [{token} {name}]"))
            else:
                result.append((f"class:{style_class}.{_type}", f" {token} {name}"))
        else:
            if is_option:
                result.append((f"class:{style_class}.{_type}", f" [{name}]"))
            else:
                result.append((f"class:{style_class}.{_type}", f" {name}"))
    elif isinstance(name, list):
        for inner_name, inner_type in zip(name, _type):
            inner_type = type_convert.get(inner_type, inner_type)
            if is_option:
                result.append((f"class:{style_class}.{inner_type}", f" [{inner_name}]"))
            else:
                result.append((f"class:{style_class}.{inner_type}", f" {inner_name}"))
    else:
        raise Exception()
    return result


def compose_command_syntax(command_info, style_class="bottom-toolbar"):
    command_style = f"class:{style_class}.command"
    const_style = f"class:{style_class}.const"
    args = []
    if command_info.get("arguments"):
        for argument in command_info["arguments"]:
            if argument.get("command"):
                # command [
                args.append((command_style, " [" + argument["command"]))
                if argument.get("enum"):
                    enums = "|".join(argument["enum"])
                    args.append((const_style, f" [{enums}]"))
                elif argument.get("name"):
                    args.extend(
                        parse_argument_to_formatted_text(
                            argument["name"],
                            argument["type"],
                            argument.get("optional"),
                            token=argument.get("token"),
                            style_class=style_class,
                        )
                    )
                # ]
                args.append((command_style, "]"))
            elif argument.get("enum"):
                enums = "|".join(argument["enum"])
                args.append((const_style, f" [{enums}]"))

            else:
                args.extend(
                    parse_argument_to_formatted_text(
                        argument["name"],
                        argument["type"],
                        argument.get("optional"),
                        token=argument.get("token"),
                        style_class=style_class,
                    )
                )
    return args


def command_syntax(command, command_info):
    """
    Get command syntax based on redis-doc/commands.json

    :param command: Command name in uppercase
    :param command_info: dict loaded from commands.json, only for
        this command.
    """
    comamnd_group = command_info["group"]
    bottoms = [
        ("class:bottom-toolbar.group", f"({comamnd_group}) "),
        ("class:bottom-toolbar.command", f"{command}"),
    ]  # final display FormattedText

    bottoms += compose_command_syntax(command_info)

    if "since" in command_info:
        since = command_info["since"]
        bottoms.append(("class:bottom-toolbar.since", f"   since: {since}"))
    if "complexity" in command_info:
        complexity = command_info["complexity"]
        bottoms.append(("class:bottom-toolbar.complexity", f" complexity:{complexity}"))

    return FormattedText(bottoms)


def _literal_bytes(b):
    """
    convert bytes to printable text.

    backslash and double-quotes will be escaped by
    backslash.
    "hello\" -> \"hello\\\"

    we don't add outer double quotes here, since
    completer also need this function's return value
    to patch completers.

    b'hello' -> "hello"
    b'double"quotes"' -> "double\"quotes\""
    """
    s = str(b)
    s = s[2:-1]  # remove b' '
    # unescape single quote
    s = s.replace(r"\'", "'")
    return s


def ensure_str(origin, decode=None):
    """
    Ensure is string, for display and completion.

    Then add double quotes

    Note: this method do not handle nil, make sure check (nil)
          out of this method.
    """
    if origin is None:
        return None
    if isinstance(origin, str):
        return origin
    if isinstance(origin, int):
        return str(origin)
    elif isinstance(origin, list):
        return [ensure_str(b) for b in origin]
    elif isinstance(origin, bytes):
        if decode:
            return origin.decode(decode)
        return _literal_bytes(origin)
    else:
        raise Exception(f"Unknown type: {type(origin)}, origin: {origin}")


def double_quotes(unquoted):
    """
    Display String like redis-cli.
    escape inner double quotes.
    add outer double quotes.

    :param unquoted: list, or str
    """
    if isinstance(unquoted, str):
        # escape double quote
        escaped = unquoted.replace('"', '\\"')
        return f'"{escaped}"'  # add outer double quotes
    elif isinstance(unquoted, list):
        return [double_quotes(item) for item in unquoted]


def exit():
    """
    Exit IRedis REPL
    """
    print("Goodbye!")
    sys.exit()


def convert_formatted_text_to_bytes(formatted_text):
    to_render = [text for style, text in formatted_text]
    return "".join(to_render).encode()


DSN = namedtuple("DSN", "scheme host port path db username password verify_ssl")


def parse_url(url, db=0):
    """
    Return a Redis client object configured from the given URL

    For example::

        redis://[[username]:[password]]@localhost:6379/0
        rediss://[[username]:[password]]@localhost:6379/0?ssl_cert_reqs=none
        unix://[[username]:[password]]@/path/to/socket.sock?db=0

    Three URL schemes are supported:

    - ```redis://``
      <http://www.iana.org/assignments/uri-schemes/prov/redis>`_ creates a
      normal TCP socket connection
    - ```rediss://``
      <http://www.iana.org/assignments/uri-schemes/prov/rediss>`_ creates a
      SSL wrapped TCP socket connection
    - ``unix://`` creates a Unix Domain Socket connection

    There are several ways to specify a database number. The parse function
    will return the first specified option:
        1. A ``db`` querystring option, e.g. redis://localhost?db=0
        2. If using the redis:// scheme, the path argument of the url, e.g.
           redis://localhost/0
        3. The ``db`` argument to this function.

    If none of these options are specified, db=0 is used.
    """
    url = urlparse(url)

    scheme = url.scheme
    path = unquote(url.path) if url.path else None
    verify_ssl = None
    # We only support redis://, rediss:// and unix:// schemes.
    # if scheme is ``unix``, read ``db`` from query string
    # otherwise read ``db`` from path
    if url.scheme == "unix":
        qs = parse_qs(url.query)
        if "db" in qs:
            db = int(qs["db"][0] or db)
    elif url.scheme in ("redis", "rediss"):
        scheme = url.scheme
        if path:
            try:
                db = int(path.replace("/", ""))
                path = None
            except (AttributeError, ValueError):
                pass
        qs = parse_qs(url.query)
        if "ssl_cert_reqs" in qs:
            verify_ssl = qs["ssl_cert_reqs"][0]
            if verify_ssl not in ["none", "optional", "required"]:
                raise ValueError(
                    f"ssl_cert_reqs must be one of 'none', 'optional', 'required' or must be omitted: {verify_ssl}"
                )
    else:
        valid_schemes = ", ".join(("redis://", "rediss://", "unix://"))
        raise ValueError(
            "Redis URL must specify one of the following" "schemes (%s)" % valid_schemes
        )

    username = unquote(url.username) if url.username else None
    password = unquote(url.password) if url.password else None
    hostname = unquote(url.hostname) if url.hostname else None
    port = url.port

    return DSN(scheme, hostname, port, path, db, username, password, verify_ssl)
